(* Delphi-Unit (Unicode)
   Copy, pack and unpack files, create and read zip archives
   optional with encryption

    Dr. J. Rathlev, D-24222 Schwentinental (info(a)rathlev-home.de)

   Acknowledgements:
     ZLibEx (Vers. 1.2.3) and ZLib routines from http://www.base2ti.com
     AES functions from http://fp.gladman.plus.com/index.htm

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Sept. 2005
   Vers. 2 - Oct. 2005   : Zip64 extensions
   Vers. 3 - Dec. 2005   : AE-2 encryption added (see AesLib)
   Vers. 4 - Apr. 2006   : Verify option added
   Vers. 4.1 - Jul. 2006 : some minor changes and corrections
   Vers. 4.2 - Aug. 2006 : new error: path too long, TZip.Destroy fixed (path>255)
   Vers. 4.3 - Sep. 2006 : exception in verify routines fixed
   Vers. 4.4 - Apr. 2007 : Warnings on setting attributes after copy
                           New error messages
   Vers. 4.5 - May  2007 : Verify on zip archives
   Vers. 4.6 - Jul. 2007 : allow setting of encryption depth on Zip files
   Vers. 4.7 - Oct. 2007 : alternate copy using SHFileOperation
                           without compression and encryption
   Vers. 4.8 - Jan. 2008 : full timestamp copy
   
   Vers. 6  - Feb. 2009 : Unicode Version for Delphi 2009
                          ===============================
     Notes: 1. GZip: The most recent file format description (4.3) comes from 1996 (RFC1952).
                     There is no Unicode or UTF-8 support specified. The actual
                     linux version of GZip writes filenames as UTF-8, but dows
                     not use any flag as marker.
                     For compatibilty this unit writes filenames in two
                     different ways:
                     a) Filename is ISO-8859 compliant: as RFC1952
                        Op.Syst. is DOS (0), filename is AnsiString
                     b) Filename is not ISO8859:
                        Op.Syst. is HTFS (11), filename is UTF8String
                     Reading files assumes AnsiString if Op.Syst. DOS, UTF8String
                     in all other cases
            2. Zip:  see PkZip App-Note Version: 6.3.2 (Sept. 2007)
                     uses general purpose flag bit 11
            3. UNC:  all full file- and directory paths are expanded to UNC
   
   Vers. 6.1   - Jul. 2009 : Read zip files, Copy file storage
   Vers. 6.2   - Jan. 2011 : several minor changes and enhancements
   Vers. 6.3   - Sep. 2011 : optional setting of filename in gz header
   Vers. 6.4   - Oct. 2011 : new thread (TVerifyThread) to compare files
                             (plain, gz packed and/or encrypted)
                             New extension for gz-encrypted files
   Vers. 6.5   - Nov. 2011 : compressed and uncompressed size as int64 in 
                             gzip extrafield
   Vers. 6.6   - Aug. 2012 : changes to use zlib routines delivered with Delph XE2
   Vers. 6.7   - Jul. 2013 : changes in reading gz header and trailer
                             optional copying of ACL, CopyOptions instead of booleans
   Vers. 6.8   - Mar. 2014 : errors in code tables OEM <> ANSI fixed
                             automatic detection of crypting length in zip files
   *)

unit ExtFileTools;

interface

uses Windows, Classes, SysUtils, ZLibex, AesLib, FtGlobal, Contnrs, FileUtils;

const
  GzExt = '.gz';      // default extension for GZip
  GzeExt = '.gze';    // extension for encrypted GZip
  GzExtMask = '.gz*';
  ZipExt = '.zip';

  (* Format of encrypted gz-files:
     Standard GZIP header,  bit 5 is set in flag byte
     Filename
     Encrypted data
     CRC32 set to 0
     Original size mod 2^32
  *)
  CrExt = '.enc';      // default extension for not compressed encrypted files
  defBufferSize = 512*1024;     // default buffer size for copy operations

  // ZLib parameters
  defWindowBits = -15;  // for GZip format
  defMemLevel = 8;

  // gz signature
  GzSignatur = $8b1f;
  Z_DEFLATE = 8;
  Z_LZMA = 14;   // LZMA

  // gzip operating system
  OsDos  = 0;
  OsUnix = 3;
  OsNTFS = 11;

  // gzip flag byte
  gfASCII      = $01; { bit 0 set: file probably ascii text }
  gfHeaderCrc  = $02; { bit 1 set: header CRC present }
  gfExtraField = $04; { bit 2 set: extra field present }
  gfFileName   = $08; { bit 3 set: original file name present }
  gfComment    = $10; { bit 4 set: file comment present }
  gfEncrypted  = $20; { bit 5 set: AE-2 encryption used for data after compressing }
  gfReserved   = $C0; { bits 6..7: reserved }

  ExtraGz64IDA = 1;         // Extra field tag for 64 size info (alternative value)
  ExtraGz64ID  = $464C;     // Extra field tag for 64 size info

  // PkZip signatures
  PkLocalSignatur = $04034b50;
  PkDirSignatur = $02014b50;
  PkEndSignatur = $06054b50;
  Pk64RecSignatur = $06064b50;
  Pk64LocSignatur = $07064b50;

  // PkZip Constants
  VersDef = $14;          // Unicode requires WinZip Vers. 12
  VersZip64 = $2D;
  VersDos = 0;
  zipEncryptFlag = 1;     // File is encrypted
  zipMaxDeflFlag = 2;     // Deflating - Maximum (-exx/-ex) compression option was used.
  zipUtf8Flag    = $800;  // Filename and comment are UTF-8 coded
  zipNoComp      = 0;     // File is stored
  zipDeflate     = 8;     // File is deflated
  zipLzma        = 14;    // Use LZMA compression
  zipEncrypted   = 99;    // Compression methode uses encryption
  ExtraZip64 = 1;         // Extra field tag for Zip64
  ExtraCrypt = $9901;     // Extra field tag for encryption
  VendorID   = $4541;     // 'AE' - see: http://www.winzip.de/aes_info.htm
  VendorVersAE1 = 1;
  VendorVersAE2 = 2;
  AE1Size       = 20;     // see: http://www.winzip.de/aes_info.htm#crc-faq
  AE2Mode128 = 1;
  AE2Mode192 = 2;
  AE2Mode256 = 3;

  tmpGzipName = 'gzip.tmp';
  tmpEncName  = 'encrypt.tmp';
  tmpUzipName = 'unzip.tmp';
  tmpCompName = 'compare.tmp';
  UMaxMemSize = 50*1024*1024;  // max. size of temp. memory stream (uncompressed)
  CMaxMemSize = 1024*1024;  // max. size of temp. memory stream (compressed)
  MaxCardinal = $FFFFFFFF;    // max. value for cardinals (4 bytes)

type
  TCopyOption = (coCompress,coZipCompress,coVerify,coOverwrite,coStorage,coPermissions,
                 coAttributes,coTimeStamp,coCreationTime,coCheckTime);
  TCopyOptions = set of TCopyOption;

  TGzHeader = packed record
    Signature : word;
    Methode,
    Flags     : byte;
    TimeStamp : cardinal;
    XFlags,
    OpSys     : byte;
    end;

  TGzExtra64 = packed record
    SubfieldID,
    Size        : word;
    CSize,USize : int64;
    end;

  TGzFileInfo = record
    Filename  : string;          // original filename
    UnixTime  : cardinal;        // original timestamp as Unix time
    Attr      : word;            // original file attributes
    CSize,
    USize     : int64;           // original size
    Encrypted : boolean;         // file is encrypted
    end;

  TExtraFieldInfo = packed record
    Tag,ESize   : word;
    end;

  TPKExtraLocalField = packed record
    Tag,ESize : word;
    Buffer    : array [0..3] of int64;
    end;

  TPKExtraEncField = packed record
    Tag,ESize,
    Version,Vendor  : word;
    Mode            : byte;
    Method          : word;
    end;

  TPKLocalHeader = packed record
    Signatur    : cardinal;
    Vers,Flag,
    Method      : word;
    FTimeStamp,
    CRC,
    CSize,USize : cardinal;
    FNLength,
    ExtraLength : word;
    end;

  TPKDirHeader = packed record
    Signatur    : cardinal;
    VersMade,
    VersExtr,
    Flag,
    Method      : word;
    FTimeStamp,
    CRC,
    CSize,USize : cardinal;
    FNLength,
    ExtraLength,
    CommLength,
    DiskStart,
    IntAttr     : word;
    ExtAttr,
    Offset      : cardinal;
    end;

  TPKEndHeader = packed record
    Signatur    : cardinal;
    ThisDisk,
    StartDisk,
    ThisEntries,
    TotalEntries : word;
    DirSize,
    Offset      : cardinal;
    CommLength  : word;
    end;

  TPKZip64CdRecord = packed record
    Signatur    : cardinal;
    Size        : int64;
    VersMade,
    VersExtr    : word;
    ThisDisk,
    StartDisk   : cardinal;
    ThisEntries,
    TotalEntries,
    DirSize,
    Offset      : int64;
    end;

  TPKZip64CdLocator = packed record
    Signatur    : cardinal;
    StartDisk   : cardinal;
    Offset      : int64;
    TotalDisk   : cardinal;
    end;

  TZipFileInfo = record
    Filename        : string;
    Flag,Method     : word;
    TimeStamp,Crc   : cardinal;
    CSize,USize     : int64;
    end;

  TPkFileRef = class(TObject)
    Offset  : int64;
    constructor Create (Ofs : int64);
    end;

  EZipError = class(EInOutError)
  public
    ZipError : integer;
    constructor Create (const ErrString : string; ErrNR : integer);
    end;

  TPkHeader = class(TObject)
    FName           : RawByteString;   // Original filename (OEM or UTF8)
	  ErrCode,
    TimeStamp,CRC   : cardinal;
    GpFlag,Attr,
    CompMethod,
    Tag,DirLength,
    EncMode,
    EncVers,EncMeth : word;
    Offset,
    CSize,USize     : int64;
    IsDir,IsAE2,
    Zip64           : boolean;
    constructor Create (const Fn : RawByteString; Ec,Al,Ts,Chk : cardinal;
                        AA,AF,AM,EV,ED,EM : word; Ofs,CSz,USz : int64);
    constructor Assign (APkHeader : TPkHeader; Ofs : int64);
    end;

const
  // Extra field for AES encryption (see: http://www.winzip.de/aes_info.htm)
  defPkExtraCrypt : TPKExtraEncField = (Tag : ExtraCrypt; ESize : 7; Version : VendorVersAE2;
                       Vendor : VendorID; Mode : AE2Mode128; Method : zipDeflate);

type
  TShellCopyMode = (scmCopy,scmQuiet);
  TShellCopyModes = set of TShellCopyMode;

  // base class for all copy functions (compressed and crypted)
  TCopyThread = class (TThread)
  private
    FOnProgress      : TProgressEvent;
    FCheckTime       : boolean;           // check for illegal filetime
    FShellCopy       : TShellCopyModes;   // use shell copy function, overrides compression and encryption
    function GetDone : boolean;
    procedure SetProgress (ACallBack : TProgressEvent);
    procedure SetTempDirectory (const ATempDir : string);
  protected
    FError           : integer;
    FBuffer          : array of byte;
    SourceName,
    DestName,
    fTmpDir          : string;
    sSource,sDest    : TStream;
    FBufSize         : cardinal;
    FUserBreak,
    FCopyTimeStamp,
    FCopyAttr,
    FCopyCrTm,
    FCopyStorage,
    FCopyAcl,
    FOverwrite,
    FVerify          : boolean;
    FErrFlag         : integer;
    FCompLevel       : TZCompressionLevel;
    FCompType        : TZStrategy;
    FCrc             : cardinal;
    UTotal,CTotal    : int64;
    FCrypting        : TEncryption;
    FAction          : TFileAction;
    FCount           : int64;
    FSpeedLimit      : boolean;
    function CopyStream (fSource,fDest : TStream; SLength : int64; ASpeedLimit : boolean = false) : integer;
    function CompareStreams (fSource,fDest : TStream; ASpeedLimit : boolean = false) : integer;
    function GzCompress (fSource,fDest : TStream; ASpeedLimit : boolean = false) : integer;
    function GzUnCompress (fSource,fDest : TStream; ASpeedLimit : boolean = false) : integer; overload;
    function GzUnCompress (fSource,fDest : TStream; SLength : int64; ASpeedLimit : boolean = false) : integer; overload;
    function EncryptFile : integer;
    function DecryptFile : integer;
    procedure EndThread (Error : integer); virtual;
    procedure UpdateProgress;
    procedure DoProgress (AAction : TFileAction; ACount : int64; ASpeedLimit : boolean = false); virtual;
    function OpenFiles (var FTime : TFileTimestamps; var Attr : word) : integer;
    function CloseFiles (ACopyStorage,ACopyAcl : boolean; FTime : TFileTimestamps; Attr : word) : integer;
    function CopyFile : integer;
    function ShCopyFile : integer;
    function VerifyFile : integer;
    function VerifyEncFile : integer;
    function CopyFileStorage : integer;
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = dest. file
    constructor Create (const ASourceName,ADestName : string;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
    destructor Destroy; override;
    procedure CancelThread; virtual;
    procedure ReleaseStreams;
    property CheckTime : boolean read FCheckTime write FCheckTime;
    property CopyAttr: boolean read FCopyAttr write FCopyAttr;
    property CopyCreationTime : boolean read FCopyCrTm write FCopyCrTm;
    property CopyPermissions : boolean read FCopyAcl write FCopyAcl;
    property CopyStorage : boolean read FCopyStorage write FCopyStorage;
    property CopyTimeStamp : boolean read FCopyTimeStamp write FCopyTimeStamp;
    property Done  : boolean read GetDone;
    property ErrorType : integer read FErrFlag;
    property ErrorCode : integer read FError;
    property ShellCopy : TShellCopyModes read FShellCopy write FShellCopy;
    property OnProgress : TProgressEvent read FOnProgress write SetProgress;
    property Overwrite : boolean read FOverwrite write FOverwrite;
    property TempDir : string read fTmpDir write SetTempDirectory;
    end;

  // Copy file with opt. encryption
  TEncryptThread = class (TCopyThread)
  protected
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = dest file,
  // AEncryption = Encrypting object (see AesLib)
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
    end;

  // Decrypt file
  TDecryptThread = class (TCopyThread)
  protected
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = dest file,
  // AEncryption = Encrypting object (see AesLib)
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
    end;

  // gzip compatible compression
  TGZipThread = class (TCopyThread)
  private
    FCompress  : boolean;
  protected
    FName : string;
    function GzWriteHeader (fDest : TStream; fName : string;
                            fTime : cardinal; fEnc : boolean) : integer;
    function GzWriteTrailer (fDest : TStream; ACrc : cardinal) : integer;
    function CompressFile : integer;
    function VerifyGzFile : integer;
    function ProcessStream (sSource : TStream): integer;
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = gz-file,
  // AEncryption = encrypting object (see AesLib), set to nil for plain compressing
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
    property CompressionLevel : TZCompressionLevel read FCompLevel write FCompLevel;
    property CompressionType : TZStrategy read FCompType write FCompType;
    property Crc : cardinal read FCrc;
    property Filename : string read FName write FName;
    end;

  // copy file with opt. compression and/or encryption
  TWinCopyThread = class (TGZipThread)
  private
    FZipCompress : boolean;
  protected
    function CompressAsZip (ACompress : boolean) : integer;
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = dest file,
  // ACompress    = use gzip compression
  // AVerify      = verify after operation
  // AEncryption  = encrypting object (see AesLib), set to nil for no encryption
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        AOptions : TCopyOptions; ASuspend : Boolean; ABufSize : integer;
                        APriority : TThreadPriority);
    end;

  // gzip compatible uncompression
  TGUnZipThread = class (TCopyThread)
  protected
    function GzReadHeader (fSource : TStream; var fName : string;
                           var fTime : cardinal; var fEnc  : boolean;
                           var CSize,USize : int64) : integer;
    function GzSkipHeader (fSource : TStream) : integer;
    function GzReadSizes (fSource   : TStream; var CSize,USize : int64) : integer;
    function GzReadTrailer (fSource : TStream; USize : int64) : integer;
    function UnCompressFile : integer;
    procedure Execute; override;
  public
  // ASourceName  = gz-file,  ADestFolder = destination folder
  // ADecryption = Decrypting object (see AesLib), set to nil for plain uncompressing
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
    property Crc : cardinal read FCrc;
    end;

  // compare files
  TCompareThread = class (TGUnZipThread)
  private
    FCompress : boolean;
  protected
    function CompareDecryptedFile (fSource,fDest : TStream) : integer;
    function CompareCompressedFile(fSource,fDest : TStream) : integer;
    procedure Execute; override;
  public
    constructor Create (const AVerifyName,ARefName : string;  AEncryption : TEncryption;
                        ACompress,ASuspend : Boolean; ABufSize : integer;
                        APriority : TThreadPriority);
    end;

  // Zip compatible compression with optional encryption
  TZipWriteThread = class (TGZipThread)
  private
    FCompress,
    FSpeedLimit : boolean;
  protected
    function AddToZip : integer;
    procedure Execute; override;
  public
    constructor Create (const AFileName : string; AEncryption : TEncryption;
                        ZStream : TStream; ACompress,ASuspend : Boolean; ASpeedLimit : boolean;
                        ABufSize : integer; APriority : TThreadPriority;
                        AOnProgress : TProgressEvent); overload;
    constructor Create (ASource : TStream; AEncryption : TEncryption;
                        ZStream : TStream; ACompress,ASuspend : Boolean; ASpeedLimit : boolean;
                        ABufSize : integer; APriority : TThreadPriority;
                        AOnProgress : TProgressEvent); overload;
    end;

  // Zip compatible decompression with optional encryption for verification
  TZipVerifyThread = class (TGZipThread)
  private
    FLength : int64;
    FCompress,
    FSpeedLimit : boolean;
  protected
    function VerifyZippedFile : integer;
    procedure Execute; override;
  public
    constructor Create (AEncryption : TEncryption; ACrc : cardinal; ZStream : TStream;
                        ALength : int64; ACompress,ASuspend : Boolean; ASpeedLimit : boolean;
                        ABufSize : integer; APriority : TThreadPriority;
                        AOnProgress : TProgressEvent);
    end;

  // Zip compatible decompression with optional encryption
  // write to file or stream
  TZipReadThread = class (TGUnZipThread)
  private
    FSpeedLimit : boolean;
  protected
    function ReadZipLocalHeader(var ZipInfo : TZipFileInfo) : boolean;
    function CopyToStream (const ZipInfo : TZipFileInfo; fDest : TStream) : integer;
    function CopyToFile : integer;
    procedure Execute; override;
  public
    constructor Create (const AFileName : string; AEncryption : TEncryption;
                        ZStream : TStream; ASuspend : Boolean; ASpeedLimit : boolean;
                        ABufSize : integer; APriority : TThreadPriority;
                        AOnProgress : TProgressEvent); overload;
    constructor Create (ADest : TStream; AEncryption : TEncryption;
                        ZStream : TStream; ASuspend : Boolean; ASpeedLimit : boolean;
                        ABufSize : integer; APriority : TThreadPriority;
                        AOnProgress : TProgressEvent); overload;
    end;

  TZipCompareThread = class (TZipReadThread)
  protected
    function CompareToFile : integer;
    procedure Execute; override;
    end;

  TZip = class (TFileStream)
  private
    FKeyMode,
    FBufferSize : integer;
    FIsTemp,
    FEncrypt,
    FErrorFlag  : boolean;
    FEncryption : TEncryption;
    PkExtraCrypt : TPKExtraEncField;
    FOnProgress : TProgressEvent;
    FOnTimeout : TTimeOutEvent;
    procedure SetProgress (ACallBack : TProgressEvent);
  protected
    FPriority   : TThreadPriority;
    PBase,ZName : string;
    FileList    : TStringList;
  public
    constructor Create (const ZipName : string; const Password : AnsiString;
                        AKeyMode,ABufSize,AMode : integer);
    destructor Destroy; override;
    procedure InitEncryption (const Password : AnsiString; AKeyMode : integer);
    procedure CopyToStream(AStream : TStream; SLength : int64);
    property OnProgress : TProgressEvent read FOnProgress write SetProgress;
    property OnTimeout : TTimeOutEvent read FOnTimeout write FOnTimeout;
    property ErrorFlag : boolean read FErrorFlag write FErrorFlag;
    property IsTemp : boolean read FIsTemp write FIsTemp;
    property ThreadPriority : TThreadPriority read FPriority write FPriority;
    end;

  TZipRead = class (TZip)
  private
    ZipThread   : TGUnZipThread;
    bod : int64;
  public
    constructor Create (const ZipName : string; const Password : AnsiString;
                        AKeyMode : integer = 0; ABufSize : integer = defBufferSize);
    destructor Destroy; override;
    property Directory : TStringList read FileList;
    property DirectoryOffset : int64 read bod;
    procedure ReadDirectory;
    function Filename(Index : integer) :string;
    function IsEncrypted(Index : integer) : boolean;
    procedure ExtractToFile(Offset : int64; TimeStamp,Attr : integer; const DestFilename : string); overload;
    procedure ExtractToFile(Index : integer; const DestFilename : string); overload;
    function ExtractToStream(const Filename : string; sDest : TStream) : boolean;
    procedure CompareToFile(Offset : int64; const RefFilename : string);
    procedure Cancel;
    end;

  TZipWrite = class (TZip)
  private
    FOnVerifyProgress : TProgressEvent;
    ZipThread   : TGZipThread;
    function VerifyFile (PkHeader : TPkHeader) : boolean;
  public
    constructor Create (const ZipName,BasicDirectory : string; const Password : AnsiString;
                        AKeyMode : integer = 0; ABufSize : integer = defBufferSize);
    destructor Destroy; override;
    procedure AddDir (DirName : string);
    procedure RemLastDir;
    procedure AddStream (ASource : TStream; const AName : string; ACompress : boolean = true);
    procedure AddFile (SourceName : string; ACompress : boolean = true);
    procedure WriteDirectory;
    property Directory : TStringList read FileList write FileList;
    procedure Verify;
    procedure Suspend;
    procedure Resume;
    procedure Cancel;
    property OnVerifyProgress : TProgressEvent read FOnVerifyProgress write FOnVerifyProgress;
    end;

// retrieve gz file info
function GetGzInfo (sGz : TStream; var FileInfo : TGzFileInfo) : boolean;
function GzFileInfo (const Filename: string; var FileInfo : TGzFileInfo) : boolean;
function GzTrailer (const Filename: string; var CRC,USize : cardinal) : boolean;

// check file extension for zip, gz or enc
function IsZipFile (const Filename : string) : boolean;
function IsGzFile (const Filename : string) : boolean;
function IsGzeFile (const Filename : string) : boolean;
function IsEncFile (const Filename : string) : boolean;

implementation

uses Forms, UnicodeUtils, FilePropUtils, WinApiUtils, WinShell, HTTPApp;

const
  MaxSize = $FF000000;     // max. file size for standard zip

{ ---------------------------------------------------------------- }
// convert from ISO-8859-1 (ANSI) to Code page 437 (OEM)
// characters not found in the conversion table are to $20 = Space
type
  TCvTable = array [0..127] of AnsiChar;

  TWord = record
    case integer of
    0 : (Both : word);
    1 : (Lo,Hi : byte);
    end;

const
  TabToOEM : TCvTable =(
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,   //$80: control char.
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,   //$90
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$AD,#$9B,#$9C,#$20,#$9D,#$20,#$15,   //$A0
    #$20,#$20,#$20,#$AE,#$AA,#$20,#$20,#$20,
    #$F8,#$F1,#$FD,#$20,#$20,#$E6,#$14,#$20,   //$B0
    #$20,#$20,#$20,#$AF,#$AC,#$AB,#$20,#$A8,
    #$20,#$20,#$20,#$20,#$8E,#$8F,#$92,#$80,   //$C0
    #$20,#$90,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$A5,#$20,#$20,#$20,#$20,#$99,#$20,   //$D0
    #$20,#$20,#$20,#$20,#$9A,#$20,#$20,#$E1,
    #$85,#$20,#$83,#$20,#$84,#$86,#$91,#$87,   //$E0
    #$8A,#$82,#$88,#$89,#$8D,#$A1,#$8C,#$8B,
    #$20,#$A4,#$95,#$A2,#$93,#$20,#$94,#$F6,   //$F0
    #$20,#$97,#$A3,#$96,#$81,#$20,#$20,#$98);

  FILE_WRITE_ATTRIBUTES = $0100;

// Check if this Unicode string can be converted to OEM
function IsOEMStr (ws : string) : boolean;
var
  i : integer;
begin
  Result:=false;
  // check for Unicode characters
  for i:=0 to length(ws) do if TWord(ws[i]).Hi<>0 then exit;
  // check for ANSI characters not supported in code page 437
  for i:=0 to length(ws) do
    if (TWord(ws[i]).Lo>=$80) and (TabToOEM[TWord(ws[i]).Lo-128]=#$20) then exit;
  Result:=true;
  end;

// convert ANSI string to OEM - required for Zip
function StrToOEM (s : AnsiString) : AnsiString;
var
  i  : integer;
begin
  for i:=1 to length(s) do begin
    if s[i]>=#128 then s[i]:=TabToOEM[ord(s[i])-128];
    end;
  StrToOEM:=s;
  end;

{ ---------------------------------------------------------------- }
// convert from Code page 437 (OEM) to ISO-8859-1 (ANSI)
// characters not found in the conversion table are to $20 = Space
const
  TabToANSI : TCvTable =(
    #$C7,#$FC,#$E9,#$E2,#$E4,#$E0,#$E5,#$E7,   //$80
    #$EA,#$EB,#$E8,#$EF,#$EE,#$EC,#$C4,#$C5,
    #$C9,#$E6,#$C6,#$F4,#$F6,#$F2,#$FB,#$F9,   //$90
    #$FF,#$D6,#$DC,#$A2,#$A3,#$A5,#$20,#$20,
    #$E1,#$ED,#$F3,#$FA,#$F1,#$D1,#$20,#$20,   //$A0
    #$BF,#$20,#$AC,#$BD,#$BC,#$A1,#$AB,#$BB,
    #$20,#$20,#$20,#$A6,#$20,#$20,#$20,#$20,   //$B0
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,   //$C0
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,   //$D0
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$DF,#$20,#$20,#$20,#$20,#$B5,#$20,   //$E0
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$B1,#$20,#$20,#$20,#$20,#$F7,#$20,   //$F0
    #$B0,#$20,#$B7,#$20,#$20,#$B2,#$20,#$A0);

    (* Umwandeln eines OEM-Strings nach ANSI *)
function StrToAnsi (const s : AnsiString) : AnsiString;
var
  i  : integer;
begin
  Result:=s;
  for i:=1 to length(Result) do begin
    if Result[i]>=#128 then Result[i]:=TabToANSI[ord(Result[i])-128]
    else if Result[i]=#$15 then Result[i]:=#$A7
    else if Result[i]=#$14 then Result[i]:=#$B6;
    end;
  end;

{ ------------------------------------------------------------------- }
function GetGzHeader (fSource   : TStream;
                      var fName : string;
                      var fTime : cardinal;
                      var fEnc  : boolean;
                      var CSize,USize : int64) : integer;
var
  Header : TGzHeader;
  n,k    : word;
  b      : byte;
  s      : RawByteString;
  fp     : int64;
begin
  Result:=errOK;       // o.k.
  fName:=''; fTime:=0;
  try
    fSource.Read(Header,sizeof(TGzHeader));
    with Header do begin
      if Signature<>GzSignatur then begin
        Result:=errFileGZip; exit;
        end;
      fTime:=TimeStamp;
      with fSource do begin
        CSize:=0; USize:=0;
        if Flags and gfExtraField <> 0 then begin   // read extra field
          Read(n,2);              // length of extra field
          if n>=sizeof(TGzExtra64) then begin
            Read(k,sizeof(word));     // ID
            dec(n,2);
            if k=ExtraGz64IDA then begin
              k:=Read(k,sizeof(word));  // Data size (= 16)
              Read(CSize,sizeof(int64));
              Read(USize,sizeof(int64));
              dec(n,2+2*sizeof(int64));
              end
            end;
          Position:=Position+n;   // adjust stream position
          end;
        if Flags and gfFileName <> 0 then begin   // read filename
          repeat
            Read(b,1);
            if b<>0 then s:=s+AnsiChar(b);
            until b=0;
          end;
        if Flags and gfComment <> 0 then begin   // skip comment
          repeat
            Read(b,1);
            until b=0;
          end;
        if Flags and gfHeaderCrc <> 0 then Read(n,2); // skip crc16
        fEnc:=Flags and gfEncrypted<>0;
        if OpSys=OsDos then fName:=s
        else fName:=Utf8ToString(s);
        if CSize=0 then CSize:=Size-Position-8;
        if USize=0 then begin
          fp:=Position;
          Seek (-4,soFromEnd);
          Read(USize,4);
          Position:=fp;
          end;
        end;
      end;
  except
    Result:=errFileRead; exit; // error
    end;
  end;

{ ------------------------------------------------------------------- }
(* retrieve file info from stream *)
function GetGzInfo (sGz : TStream; var FileInfo : TGzFileInfo) : boolean;
begin
  with FileInfo do begin
    UnixTime:=DateTimeToUnixTime(Now);
    if GetGzHeader (sGz,Filename,UnixTime,Encrypted,CSize,USize)=errOK then Result:=true
    else Result:=false;
    end;
  end;

{ ------------------------------------------------------------------- }
(* retrieve file info from gz-file *)
function GzFileInfo (const Filename: string; var FileInfo : TGzFileInfo) : boolean;
var
  sFile : TFileStream;
begin
  Result:=false;
  try
    sFile:= TFileStream.Create(FilenameToUNC(Filename),fmOpenRead+fmShareDenyNone);
    Result:=GetGzInfo(sFile,FileInfo);
    if Result then FileInfo.Attr:=FileGetAttrUNC(Filename,false);
  finally
    sFile.Free;
    end;
  end;

function GzTrailer (const Filename: string; var CRC,USize : cardinal) : boolean;
var
  sFile      : TFileStream;
  aCrc,FSize : cardinal;
begin
  Result:=false;
  try
    sFile:= TFileStream.Create(FilenameToUNC(Filename),fmOpenRead+fmShareDenyNone);
    with sFile do begin
      Seek(-8,soEnd);
      Read(Crc,4);
      Read(USize,4);
      end;
    Result:=true;
  finally
    sFile.Free;
    end;
  end;

function IsGzFile (const Filename : string) : boolean;
begin
  Result:=AnsiLowercase(ExtractFileExt(Filename))=GzExt;
  end;

function IsGzeFile (const Filename : string) : boolean;
begin
  Result:=AnsiLowercase(ExtractFileExt(Filename))=GzeExt;
  end;

function IsEncFile (const Filename : string) : boolean;
begin
  Result:=AnsiLowercase(ExtractFileExt(Filename))=CrExt;
  end;

function IsZipFile (const Filename : string) : boolean;
begin
  Result:=AnsiLowercase(ExtractFileExt(Filename))=ZipExt;
  end;

function ZipCrc32(Crc : cardinal; const buf; len: Integer): cardinal;
begin
  Result:=Crc32 (Crc,pbyte(@buf),len);
  end;

{ ------------------------------------------------------------------- }
constructor TCopyThread.Create (const ASourceName,ADestName : string;
                                ASuspend : Boolean; ABufSize : integer;
                                APriority : TThreadPriority);
begin
  inherited Create (ASuspend);
  SourceName:=ASourceName;
  DestName:=ADestName;
  Priority:=APriority;
  FOnProgress:=nil;
  FCopyTimeStamp:=true;
  FCopyAttr:=true;
  FCheckTime:=true;
  FCopyCrTm:=false;
  FCopyStorage:=false;
  FCopyAcl:=false;
  FShellCopy:=[];
  FVerify:=false;
  FOverwrite:=false;
  FError:=errOK; FErrFlag:=errCopy;
  FBufSize:=ABufSize;
  SetLength(FBuffer,FBufSize);
  FUserBreak:=false;
  UTotal:=0; CTotal:=0;
  // init checksum
  FCrc:=Crc32(0,nil,0);
  FCompLevel:=zcDefault; FCompType:=zsDefault;
  FCrypting:=nil;
  fTmpDir:=TempDirectory;
  end;

destructor TCopyThread.Destroy;
begin
  FBuffer:=nil;
  if assigned(FCrypting) then FCrypting.OnProgress:=nil;
  inherited Destroy;
  end;

function TCopyThread.GetDone : boolean;
begin
  Result:=Terminated;
  end;

procedure TCopyThread.SetProgress (ACallBack : TProgressEvent);
begin
  FOnProgress:=ACallBack;
  if assigned(FCrypting) then FCrypting.OnProgress:=ACallBack;
  end;

procedure TCopyThread.SetTempDirectory (const ATempDir : string);
begin
  if (length(ATempDir)>0) and DirectoryExists(ATempDir) then fTmpDir:=SetDirName(ATempDir)
  else fTmpDir:=TempDirectory;
  end;

procedure TCopyThread.EndThread (Error : integer);
begin
  if Error<>errOK then FError:=FErrFlag or Error;
  Terminate;
  end;

procedure TCopyThread.CancelThread;
begin
  FUserBreak:=true;
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.CopyStream (fSource,fDest : TStream; SLength : int64; ASpeedLimit : boolean) : integer;
var
  NRead,NWrite,ec,fb : integer;
begin
  ec:=errOK;
  repeat
    if SLength<FBufSize then fb:=SLength else fb:=FBufSize;
    try
      NRead:=sSource.Read(FBuffer[0],fb);
      if NRead<fb then ec:=errError+errFileRead   // z.B. wenn "sSource" gelockt ist
      else begin
        UTotal:=UTotal+NRead;        // total number of bytes read
        DoProgress(acCopy,UTotal,ASpeedLimit);
        FCrc:=ZipCrc32(FCrc,FBuffer[0],NRead);
        end;
    except
      ec:=errError+errFileRead;
      end;
    if ec=errOK then begin
      try
        NWrite:=sDest.Write(FBuffer[0],NRead);
        if NWrite<NRead then ec:=errError+errFileFull; // Ziel-Medium voll
      except
        ec:=errError+errFileWrite;
        end;
      if FUserBreak then ec:=errUserBreak;
      end;
    dec(Slength,NRead);
    until (SLength<=0) or (ec<>errOK);
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.CompareStreams (fSource,fDest : TStream; ASpeedLimit : boolean) : integer;
var
  NRead1,NRead2,fb : integer;
  CBuffer          : array of byte;
  SLength,STotal   : int64;
begin
  sLength:=fSource.Size;
  SetLength(CBuffer,FBufSize);
  STotal:=0;
  DoProgress(acVerify,-SLength,ASpeedLimit);
  if SLength<>fDest.Size then Result:=errError+errFileVerify
  else begin
    fSource.Seek(0,soFromBeginning);
    fDest.Seek(0,soFromBeginning);
    Result:=errOK;
    repeat
      if SLength<FBufSize then fb:=SLength else fb:=FBufSize;
      try
        NRead1:=fSource.Read(FBuffer[0],fb);
        if NRead1<fb then Result:=errError+errFileRead   // z.B. wenn "sSource" gelockt ist
        else begin
          STotal:=STotal+NRead1;        // total number of bytes read
          DoProgress(acVerify,STotal,ASpeedLimit);
          end;
      except
        Result:=errError+errFileRead;
        end;
      if Result=errOK then begin
        try
          NRead2:=fDest.Read(CBuffer[0],NRead1);
          if NRead2<NRead1 then Result:=errError2+errFileRead   // z.B. wenn "sDest" gelockt ist
        except
          Result:=errError2+errFileRead;
          end;
        if FUserBreak then Result:=errUserBreak;
        end;
      dec(Slength,NRead1);
      if (Result=errOK) and not CompareMem(@FBuffer[0],@CBuffer[0],NRead1) then
        Result:=errError+errFileVerify;
      until (SLength<=0) or (Result<>errOK);
    end;
  CBuffer:=nil;
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.GzCompress (fSource,fDest : TStream; ASpeedLimit : boolean) : integer;
var
  sComp         : TZCompressionStream;
  NRead         : cardinal;
  ec            : integer;
  fp            : int64;
begin
  ec:=errOK;       // o.k.
  // setup for GZip compatible compression
  sComp:=TZCompressionStream.Create(fDest,FCompLevel,defWindowBits,defMemLevel,FCompType);
  fp:=fDest.Position;
  repeat
    try
      NRead:=fSource.Read(FBuffer[0],FBufSize);
      UTotal:=UTotal+NRead;        // total number of bytes read
      DoProgress(acCompress,UTotal,ASpeedLimit);
      FCrc:=ZipCrc32(FCrc,FBuffer[0],NRead);
    except
      ec:=errError+errFileRead;
      end;
    if ec=errOK then begin
      try
        sComp.Write(FBuffer[0],NRead);
//   does not work with TZCompressionStream: NWrite is always equal to NRead
//        if NWrite<NRead then ec:=errError+errFileFull;  // dest disk full
      except
        ec:=errError+errFileWrite;
        end;
      if (ec=errOK) and FUserBreak then ec:=errUserBreak;
      end;
    until (NRead<FBufSize) or (ec<>errOK);
  if ec=errOK then sComp.Free else sComp.Break;
  if (ec=errOK) and (fSource.Size>0) and (fDest.Position<=fp) then
    ec:=errError+errFileFull;  // dest disk full
  Result:=ec;
  end;

function TCopyThread.GzUnCompress (fSource,fDest : TStream; ASpeedLimit : boolean) : integer;
var
  ALength : int64;
begin
  with fSource do ALength:=Size-Position-8; // gz-format
  Result:=GzUnCompress (fSource,fDest,ALength,ASpeedLimit);
  end;

function TCopyThread.GzUnCompress (fSource,fDest : TStream; SLength : int64; ASpeedLimit : boolean) : integer;
var
  sComp         : TZDecompressionStream;
  NRead,NWrite  : cardinal;
  RTotal        : int64;
  ec            : integer;
begin
  ec:=errOK;       // o.k.
  // setup for GZip compatible decompression
  sComp:=TZDecompressionStream.Create(fSource,SLength,defWindowBits);
  RTotal:=0;
  repeat
    try
      NRead:=sComp.Read(FBuffer[0],FBufSize);
    except
      ec:=errError+errFileRead;
      end;
    if ec=errOK then begin
      try
        FCrc:=ZipCrc32(FCrc,FBuffer[0],NRead);
        RTotal:=RTotal+NRead;
        DoProgress(acUnCompress,RTotal,ASpeedLimit);
        if assigned(fDest) then NWrite:=fDest.Write(FBuffer[0],NRead)
        else NWrite:=NRead;
        UTotal:=UTotal+NWrite;        // total number of bytes written
        if (NWrite<NRead) then ec:=errError+errFileFull;    // dest disk full
      except
        ec:=errError+errFileWrite;
        end;
      if (ec=errOK) and FUserBreak then ec:=errUserBreak;
      end;
    until (NRead<FBufSize) or (ec<>errOK);;
  sComp.Free;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
procedure TCopyThread.DoProgress (AAction : TFileAction; ACount : int64; ASpeedLimit : boolean);
begin
  if Assigned(FOnProgress) then begin
    FAction:=AAction; FCount:=ACount; FSpeedLimit:=ASpeedLimit;
    Synchronize(UpdateProgress);
    end;
  end;

procedure TCopyThread.UpdateProgress;
begin
  FOnProgress(FAction,FCount,FSpeedLimit);
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.OpenFiles (var FTime : TFileTimestamps; var Attr : word) : integer;
begin
  Result:=errOK;
//  if length(SourceName)>MAX_PATH-1 then Result:=errError+errLongPath else
  if not FileExistsUNC(SourceName,false) then Result:=errError+errNotFound;
  if Result=errOK then begin
//    if length(DestName)>MAX_PATH-1 then Result:=errError+errLongPath else
    if (length(DestName)=0) then Result:=errError+errFileCreate;
    end;
  if Result<>errOK then exit;
  FTime:=GetFileTimestamps(SourceName,false);
  Attr:=FileGetAttrUNC(SourceName,false);
  try
    sSource:=TFileStream.Create(FilenameToUNC(SourceName),fmOpenRead+fmShareDenyNone);
  except
    on EFOpenError do begin
      Result:=errError+errFileOpen; Exit;
      end;
    end;
  // overwrite destination, reset attributes
  if FOverwrite and FileExistsUNC(DestName) then begin
    if FileSetAttrUNC(DestName,faArchive)<>0 then begin
      Result:=errError+errFileAttr; Exit;
      end;
    end;
  try
    sDest:=TFileStream.Create(FilenameToUNC(DestName),fmCreate);
  except
    on EFCreateError do Result:=errError+errFileCreate;
    end;
  end;

{ ------------------------------------------------------------------- }
procedure TCopyThread.ReleaseStreams;
begin
  if assigned(sSource) then try FreeAndNil(sSource); except end;
  if assigned(sDest) then try FreeAndNil(sDest); except end;
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.CloseFiles (ACopyStorage,ACopyAcl : boolean;
                                 FTime : TFileTimestamps; Attr : word) : integer;
begin
  Result:=errOK;
  if FVerify then begin
    if not FlushFileBuffers(TFileStream(sDest).Handle) then Result:=errFileWrite;
    end;
  ReleaseStreams;
  if FCopyTimeStamp then begin
    if SetFileTimestamps(DestName,FTime,FCheckTime,FCopyCrTm,false)>0 then
      Result:=errWarning+errFileTS;
    end;
  if (Result=errOK) and FCopyAttr then begin
    if FileSetAttrUNC(DestName,Attr,false)>0 then Result:=errWarning+errFileAttr;
    end;
  if (Result=errOK) and ACopyAcl then begin
    if CopyFileAcl(SourceName,DestName)<>ERROR_SUCCESS then Result:=errWarning+errAcl;
    end;
  if (Result=errOK) and ACopyStorage then Result:=CopyFileStorage;
  end;

{ ------------------------------------------------------------------- }
// This routine only supports pathnames < MAX_PATH (=260)
// Reason: the called function "StgOpenStorageEx" does not support UNC paths
function TCopyThread.CopyFileStorage : integer;
var
  Summary : TFileSummary;
begin
  if GetFileSummary(SourceName,Summary) then begin
    if SetFileSummary(DestName,Summary) then Result:=errOK
    else Result:=errWarning+errStorage;
    end
  else Result:=errWarning+errStorage;
  end;

{ ------------------------------------------------------------------- }
// encrypt file
function TCopyThread.EncryptFile : integer;
var
  NRead    : cardinal;
  FTime    : TFileTimestamps;
  Attr     : word;
  ec       : integer;
begin
  ec:=OpenFiles (FTime,Attr);
  if (ec=errOK) then begin
    DoProgress(acEncrypt,-sSource.Size,false);   // Reset
    with FCrypting do if WriteHeader (sDest) then begin
      repeat
        try
          NRead:=sSource.Read(FBuffer[0],FBufSize);
          UTotal:=UTotal+NRead;        // total number of bytes read
          DoProgress(acEncrypt,UTotal,true);
        except
          ec:=errError+errFileRead;
          end;
        if ec=errOK then begin
          try
            EncryptBlock (FBuffer[0],NRead);
            sDest.Write(FBuffer[0],NRead);
          except
            ec:=errError+errFileECrypt;
            end;
          if  FUserBreak then ec:=errUserBreak;
          end;
        until (NRead<FBufSize) or (ec<>errOK);
      if (ec=errOK) and not WriteTrailer (sDest) then ec:=errError+errFileECrypt;
      end
    else ec:=errError+errFileECrypt;
    if (ec=errOK) and (UTotal<>sSource.Size) then ec:=errError+errFileRead;  // z.B. wenn "sSource" gelockt ist
    if ec=errOK then ec:=CloseFiles (FCopyStorage,FCopyAcl,FTime,Attr) else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

// verify encrypted file
function TCopyThread.VerifyEncFile : integer;
var
  ec        : integer;
begin
  ec:=errOK;
  try
    sDest:=TFileStream.Create(FilenameToUNC(DestName),fmOpenRead+fmShareDenyNone);
  except
    on EFOpenError do ec:=errError+errFileVerify;
    end;
  if ec=errOK then begin
    DoProgress(acAddVerify,-UTotal,false);   // Reset progress indicator
    if not FCrypting.DecryptStream(sDest,nil,sDest.Size,true) then ec:=errError+errFileVerify;
    end;
  if assigned(sDest) then try FreeAndNil(sDest); except end;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
// decrypt file
function TCopyThread.DecryptFile : integer;
var
  FTime             : TFileTimestamps;
  Attr              : word;
  ec                : integer;
begin
  ec:=OpenFiles (FTime,Attr);
  if (ec=errOK) then begin
    if not FCrypting.DecryptStream(sSource,sDest,sSource.Size,true) then ec:=errError+errFileDCrypt;
    if ec=errOK then ec:=CloseFiles (false,FCopyAcl,FTime,Attr) else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
// copy file
function TCopyThread.CopyFile : integer;
var
  FTime             : TFileTimestamps;
  ec                : integer;
  Attr              : word;
begin
  ec:=OpenFiles (FTime,Attr);
  if ec=errOK then begin
    DoProgress(acCopy,-sSource.Size,true);
    ec:=CopyStream(sSource,sDest,sSource.Size,true);
    if ec=errOk then ec:=CloseFiles(FCopyStorage,FCopyAcl,FTime,Attr) else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

// copy using shell function "SHFileOperation" to include file infos
function TCopyThread.ShCopyFile : integer;
var
  FTime     : TFileTimestamps;
  NRead     : cardinal;
  quiet     : boolean;
begin
  Result:=errOK;
  FTime:=GetFileTimestamps(SourceName);
//  if length(SourceName)>MAX_PATH-1 then Result:=errError+errLongPath else
  if not FileExistsUNC(SourceName,false) then Result:=errError+errNotFound;
  if Result=errOK then begin
//    if length(DestName)>MAX_PATH-1 then Result:=errError+errLongPath else
    if (length(DestName)=0) then Result:=errError+errFileCreate;
    end;
  if Result=errOK then begin
    quiet:=(scmQuiet in FShellCopy) or (LongFileSize(SourceName)<CMaxMemSize);
    DoProgress(acCopy,-1,false);   // Reset progress indicator
    if not ShellCopyFiles(Application.Handle,SourceName,DestName,'',quiet) then
       Result:=errError+errFileWrite;
    end;
  if FVerify and (Result=errOK) then begin   //Checksumme ermitteln
    UTotal:=0;
    try
      sSource:=TFileStream.Create(FilenameToUNC(SourceName),fmOpenRead+fmShareDenyNone);
    except
      on EFOpenError do Result:=errError+errFileVerify;
      end;
    if Result=errOK then begin
      DoProgress(acAddVerify,-sSource.Size,false);   // Reset progress indicator
      FCrc:=Crc32(0,nil,0);
      repeat
        try
          NRead:=sSource.Read(FBuffer[0],FBufSize);
          UTotal:=UTotal+NRead;
          DoProgress(acAddVerify,UTotal,true);
          FCrc:=ZipCrc32(FCrc,FBuffer[0],NRead);
        except
          Result:=errError+errFileVerify; NRead:=0;
          end;
        if FUserBreak then Result:=errUserBreak;
        until (NRead<FBufSize) or (Result<>errOK);
      FreeAndNil(sSource);
      end;
    end;
  if (Result=errOK) then begin
    if FCopyTimeStamp and FCopyCrTm then begin
      if SetFileTimestamps(DestName,FTime,FCheckTime,FCopyCrTm,false)>0 then
        Result:=errWarning+errFileTS;
      end;
    end;
  end;

// verify check sum of copied file
function TCopyThread.VerifyFile : integer;
var
  ec        : integer;
  NRead,Crc : cardinal;
  RTotal    : int64;
begin
  ec:=errOK; RTotal:=0;
  try
    sDest:=TFileStream.Create(FilenameToUNC(DestName),fmOpenRead+fmShareDenyNone);
  except
    on EFOpenError do ec:=errError+errFileVerify;
    end;
  if ec=errOK then begin
    DoProgress(acAddVerify,-UTotal,false);   // Reset progress indicator
    Crc:=Crc32(0,nil,0);
    repeat
      try
        NRead:=sDest.Read(FBuffer[0],FBufSize);
        RTotal:=RTotal+NRead;
        DoProgress(acAddVerify,RTotal,true);
        Crc:=ZipCrc32(Crc,FBuffer[0],NRead);
      except
        ec:=errError+errFileVerify; NRead:=0;
        end;
      if FUserBreak then ec:=errUserBreak;
      until (NRead<FBufSize) or (ec<>errOK);
    if (ec=errOK) and (FCrc<>Crc) then ec:=errError+errFileVerify;
    end;
  if assigned(sDest) then try FreeAndNil(sDest); except end;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
// execute thread
procedure TCopyThread.Execute;
begin
  if scmCopy in FShellCopy then EndThread (ShCopyFile)
  else EndThread (CopyFile);
  end;

{ TEncryptThread ---------------------------------------------------------------- }
constructor TEncryptThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,ASuspend,ABufSize, APriority);
  FCrypting:=AEncryption;
  if assigned(FCrypting) then FErrFlag:=errEnCrypt;
  end;

{ ------------------------------------------------------------------- }
// execute thread
procedure TEncryptThread.Execute;
begin
  if assigned(FCrypting) then begin
//    DestName:=DestName+CrExt;
    EndThread (EncryptFile);
    end
  else EndThread (CopyFile);
  end;

{ TDecryptThread ---------------------------------------------------------------- }
constructor TDecryptThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,ASuspend,ABufSize, APriority);
  FCrypting:=AEncryption;
  if assigned(FCrypting) then FErrFlag:=errDeCrypt;
  end;

{ ------------------------------------------------------------------- }
// execute thread
procedure TDecryptThread.Execute;
begin
  if assigned(FCrypting) then EndThread (DecryptFile) else EndThread (CopyFile);
  end;

{ TGZipThread ------------------------------------------------------------------- }
constructor TGZipThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                                ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,ASuspend,ABufSize, APriority);
  FName:=ExtractFilename(SourceName);
  FCompress:=true;
  FCrypting:=AEncryption;
  FErrFlag:=errGZip;
  end;

{ ------------------------------------------------------------------- }
// write gz header
function TGZipThread.GzWriteHeader (fDest : TStream;
                                    fName : string;
                                    fTime : cardinal;
                                    fEnc  : boolean) : integer;
var
  Header : TGzHeader;
  n      : integer;
  w      : word;
  os     : byte;
  s      : RawByteString;
  ef     : TGzExtra64;
begin
  Result:=errOK;       // o.k.
  if IsAnsiStr(fname) then begin    // gzip uses ISO-8859-1
    s:=fName; os:=OsDos;
    end
  else begin
    s:=UTF8Encode(fName); os:=OsNTFS;
    end;
  n:=length(s);
  with Header do begin
    Signature:=GzSignatur;
    Methode:=Z_DEFLATED;
    Flags:=gfExtraField;
    if n>0 then Flags:=Flags or gfFileName;
    if fEnc then Flags:=Flags or gfEncrypted;
    TimeStamp:=fTime;
    XFlags:=0;
    OpSys:=os;
    end;
  try
    // write header
    if fDest.Write(Header,sizeof(TGzHeader))<sizeof(TGzHeader) then begin
      Result:=errError+errFileFull;  // dest disk full
      Exit;
      end;
    // write extra field with size information
    w:=sizeof(ef);
    fDest.Write(w,2);
    with ef do begin
      SubfieldID:=ExtraGz64IDA; Size:=2*sizeof(int64);
      CSize:=0; USize:=0;
      end;
    if fDest.Write(ef,sizeof(ef))<sizeof(ef) then begin
      Result:=errError+errFileFull;  // dest disk full
      Exit;
      end;
    if n>0 then begin   // write filename
      if fDest.Write(s[1],n+1)<n+1 then Result:=errError+errFileFull;  // dest disk full
      end;
  except
    Result:=errError+errFileWrite; // error
    end;
  end;

{ ------------------------------------------------------------------- }
// write gz trailer
function TGZipThread.GzWriteTrailer (fDest : TStream; ACrc : cardinal) : integer;
begin
  Result:=errOK;       // o.k.
  try
    with fDest do if Write(ACrc,4)<4 then Result:=errError+errFileFull  // dest disk full
    else if Write(TInt64(UTotal).Lo,4)<4 then Result:=errError+errFileFull  // dest disk full
    else begin  // write size info to extra field
      Position:=sizeof(TGzHeader)+6;
      Write(CTotal,sizeof(int64));
      Write(UTotal,sizeof(int64));
      end;
  except
    Result:=errError+errFileWrite; // error
    end;
  end;

{ ------------------------------------------------------------------- }
// compress file
function TGZipThread.CompressFile  : integer;
var
  FTime         : TFileTimestamps;
  utime         : cardinal;
  fp            : int64;
  Attr          : word;
  sTmp          : TStream;
  ec            : integer;
  FEnc,tf,sl    : boolean;
  tzn           : string;
begin
  ec:=OpenFiles (FTime,Attr);
  if ec=errOK then begin
    FEnc:=assigned(FCrypting);
    // unix time for gzip
    utime:=FileTimeToUnixTime(FTime.LastWriteTime);
    ec:=GzWriteHeader(sDest,FName,utime,FEnc);
    fp:=sDest.Position;
    if FEnc then begin
      tf:=sSource.Size>UMaxMemSize;
      if tf then begin
        tzn:=fTmpDir+tmpEncName;
        sTmp:=TFileStream.Create(FilenameToUNC(tzn),fmCreate); // write to disk
        end
      else sTmp:=TMemoryStream.Create;
      sl:=false;
      end
    else begin
      sTmp:=sDest; sl:=true;
      end;
    // compress file
    if ec=errOK then begin
      DoProgress(acCompress,-sSource.Size,false);   // Reset progress indicator
      ec:=GzCompress(sSource,sTmp,sl);
      if ec=errOK then begin
        // z.B. wenn "sSource" gelockt ist (siehe LockFile)
        if UTotal<>sSource.Size then ec:=errError+errFileRead;
        // encrypt file
        if FEnc then begin
          if (ec=errOK) then begin
            DoProgress(acEncrypt,-sTmp.Size,false);   // Reset progress indicator
            sTmp.Seek(0,soFromBeginning);
            if (ec=errOK) and not FCrypting.EncryptStream(sTmp,sDest,true) then
              ec:=errError+errFileECrypt;
            end;
          sTmp.Free;
          if tf then DeleteFile(tzn);
          end;
        if ec=errOK then begin
          CTotal:=sDest.Position-fp;
//          if UTotal>MaxCardinal then UTotal:=MaxCardinal;
          if FEnc then ec:=GzWriteTrailer (sDest,0)
          else ec:=GzWriteTrailer (sDest,FCrc);
          end;
        end;
      end;
    if ec=errOK then ec:=CloseFiles(FCopyStorage,FCopyAcl,FTime,Attr) else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

// verify check sum of copied file
function TGZipThread.VerifyGzFile : integer;
var
  fn            : string;
  ft,Crc        : cardinal;
  FEnc          : boolean;
  CSize,USize   : int64;
  ec            : integer;
begin
  ec:=errOK;
  try
    sDest:=TFileStream.Create(FilenameToUNC(DestName),fmOpenRead+fmShareDenyNone);
  except
    on EFOpenError do ec:=errError+errFileVerify;
    end;
  if ec=errOK then begin
    CrC:=FCrc;
    ec:=GetGzHeader (sDest,fn,ft,FEnc,CSize,USize);
    DoProgress(acAddVerify,-USize,false);   // Reset progress indicator
    FCrc:=Crc32(0,nil,0);
    if ec=errOK then begin
      if FEnc then begin // decrypt file
//        with sDest do n:=Size-Position-8;
        if not FCrypting.DecryptStream(sDest,nil,CSize,true) then ec:=errError+errFileVerify;
        end
      else begin
        ec:=GzUncompress(sDest,nil,true);
        if (ec=errOK) and (FCrc<>Crc) then ec:=errError+errFileVerify;
        end;
      end;
    end;
  if assigned(sDest) then try FreeAndNil(sDest); except end;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
//
function TGZipThread.ProcessStream (sSource : TStream): integer;
var
  fp     : int64;
  sTmp   : TStream;
  ec     : integer;
  tf     : boolean;
  tzn    : string;
begin
  fp:=sDest.Position;
  DoProgress(acCompress,-sSource.Size,false);   // Reset progress indicator
  if assigned(FCrypting) then begin
    if (sSource.Size>0) and FCompress then begin
      tf:=sSource.Size>UMaxMemSize;
      if tf then begin
        tzn:=fTmpDir+tmpEncName;
        sTmp:=TFileStream.Create(FilenameToUNC(tzn),fmCreate); // write to disk
        end
      else sTmp:=TMemoryStream.Create;
      ec:=GzCompress(sSource,sTmp,FSpeedLimit);  // size = 0, no compression
      if ec=errOK then begin
        // z.B. wenn "sSource" gelockt ist (siehe LockFile)
        if UTotal<>sSource.Size then ec:=errError+errFileRead;
        end;
      end
    else begin
      sTmp:=sSource; ec:=errOK; UTotal:=sSource.Size;
      end;
    DoProgress(acEncrypt,-sTmp.Size,false);   // Reset progress indicator
    // encrypt file
    if (ec=errOK) then begin
      sTmp.Seek(0,soFromBeginning);
      if not FCrypting.EncryptStream(sTmp,sDest,FSpeedLimit) then ec:=errError+errFileECrypt;
      end;
    sTmp.Free;
    if tf then DeleteFile(tzn);
    end
  else begin
    if (sSource.Size>0) and FCompress then ec:=GzCompress(sSource,sDest,FSpeedLimit)  // size = 0, no compression
    else ec:=CopyStream(sSource,sDest,sSource.Size,true);
    if ec=errOK then begin
      // z.B. wenn "sSource" gelockt ist (siehe LockFile)
      if UTotal<>sSource.Size then ec:=errError+errFileRead;
      end;
    end;
  CTotal:=sDest.Position-fp;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
// execute thread
procedure TGZipThread.Execute;
begin
  EndThread (CompressFile);
  end;

{ TWinCopyThread ------------------------------------------------------------------- }
constructor TWinCopyThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                                   AOptions : TCopyOptions; ASuspend : Boolean; ABufSize : integer;
                                   APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,AEncryption,ASuspend,ABufSize,APriority);
  FCompress:=coCompress in AOptions; FZipCompress:=coZipCompress in AOptions;
  FOverwrite:=coOverwrite in AOptions; FCopyStorage:=coStorage in AOptions;
  FCopyAttr:=coAttributes in AOptions; FCopyTimeStamp:=coTimeStamp in AOptions;;
  FCopyAcl:=coPermissions in AOptions; FCopyCrTm:=coCreationTime in AOptions;
  FVerify:=coVerify in AOptions; FCheckTime:=coCheckTime in AOptions;
  if FZipCompress then FErrFlag:=errZip
  else if assigned(FCrypting) then FErrFlag:=errEncrypt
  else if FCompress then FErrFlag:=errGZip
  else FErrFlag:=errCopy;
  end;

function TWinCopyThread.CompressAsZip (ACompress : boolean) : integer;
var
  PkExtraCrypt : TPKExtraEncField;
  ec        : integer;
  FTime     : TFileTimestamps;
  Attr      : word;
  ZfName    : RawByteString;
  fn        : string;
  utime     : cardinal;
  LFHeader  : TPKLocalHeader;
  lef       : TPKExtraLocalField;
  FileData  : TWin32FindData;
  fp,ut     : int64;
  FEnc,
  z64,zae2,
  timo      : boolean;
  rf,fc,el  : cardinal;
  pke       : TPkEndHeader;
  pkd       : TPkDirHeader;
  pkz       : TPKZip64CdRecord;
  pko       : TPKZip64CdLocator;
  ez64l,n   : integer;
begin
  ec:=OpenFiles (FTime,Attr);
  if ec=errOK then begin
    PkExtraCrypt:=defPkExtraCrypt;
    FEnc:=assigned(FCrypting);
    if GetFileData(SourceName,FileData) then begin
      utime:=GetFileAge(FileData);
      ut:=GetFileSize(FileData);
      z64:=ut>MaxSize;
      zae2:=ut<AE1Size;
      with LFHeader do begin
        Signatur:=PKLocalSignatur;
        fn:=ExtractFilename(SourceName);
        if IsOEMStr(fn) then begin
          Flag:=0; ZfName:=fn;
          ZfName:=StrToOEM(ZfName); // ZIP needs OEM characters (like DOS)
          end
        else begin
          Flag:=zipUtf8Flag;
          ZfName:=UTF8Encode(fn);
          end;
        if z64 then Vers:=VersZip64 else Vers:=VersDef;
        if FEnc then begin
          Flag:=Flag or zipEncryptFlag;
          Method:=zipEncrypted;
          ACompress:=true;         // compress always on encryption
          end
        else begin
          Flag:=Flag or zipMaxDeflFlag;
          if (ut>0) and ACompress then Method:=zipDeflate else Method:=zipNoComp;
          end;
        FTimeStamp:=utime;
        CRC:=0;
        FNLength:=length(ZfName);
        if z64 then begin
          with lef do begin
            Tag:=ExtraZip64;
            ESize:=16;
            Buffer[0]:=0;
            Buffer[1]:=0;
            end;
          CSize:=MaxCardinal;
          USize:=MaxCardinal;
          el:=20;
          ExtraLength:=20;
          end
        else begin
          CSize:=0;
          USize:=0;
          ExtraLength:=0;
          end;
        if FEnc then begin
          ExtraLength:=ExtraLength+sizeof(TPKExtraEncField);
          with PkExtraCrypt do begin
            if zae2 then Version:=VendorVersAE2 else Version:=VendorVersAE1;
            if ut>0 then Method:=zipDeflate else Method:=zipNoComp;
            end;
          end;
        end;
      with sDest do try
        Write(LFHeader,sizeof(TPKLocalHeader));         // local header
        Write(ZfName[1],LFHeader.FNLength);             // filename
        if z64 then Write(lef,el);                      // extra Zip64 field
        if FEnc then Write(PkExtraCrypt,sizeof(TPKExtraEncField));  // extra enc field
        ec:=ProcessStream(sSource);
        if ec=errOK then begin
          fc:=FCrc;
          if FEnc and zae2 then Position:=18    // position of compressed size
          else begin                            // Crc if no encryption or AE-1
            Position:=14;                       // position of crc
            Write(fc,sizeof(cardinal));
            end;
          if z64 then begin
            // write 8 byte values to extra field
            Position:=sizeof(TPKLocalHeader)+LFHeader.FNLength+4;
            Write (UTotal,sizeof(int64));
            Write (CTotal,sizeof(int64));
            end
          else begin
            // write 4 byte values
            rf:=TInt64(CTotal).Lo;
            Write(rf,sizeof(cardinal));
            rf:=TInt64(UTotal).Lo;
            Write(rf,sizeof(cardinal));
            end;
          Seek (0,soEnd);
          fp:=Position;
          // end directory
          with pkd do begin
            Signatur:=PkDirSignatur;
            Flag:=LFHeader.Flag;
            Method:=LFHeader.Method;
            FTimeStamp:=LFHeader.FTimeStamp;
            FNLength:=LFHeader.FNLength;
            if z64 then begin
              VersMade:=VersZip64+VersDos;
              VersExtr:=VersZip64+VersDos;
              with lef do begin
                Tag:=ExtraZip64;
                n:=0;
                if UTotal>MaxCardinal then begin
                  USize:=MaxCardinal;
                  Buffer[n]:=UTotal;
                  inc(n);
                  end
                else USize:=TInt64(UTotal).Lo;
                if CTotal>MaxCardinal then begin
                  CSize:=MaxCardinal;
                  Buffer[n]:=CTotal;
                  inc(n);
                  end
                else CSize:=TInt64(CTotal).Lo;
                Offset:=0;
                ESize:=8*n;
                ez64l:=4+ESize;
                ExtraLength:=ez64l;
                end;
              end
            else begin
              VersMade:=VersDef+VersDos;
              VersExtr:=VersDef;
              CSize:=TInt64(CTotal).Lo;
              USize:=TInt64(UTotal).Lo;
              Offset:=0;
              ExtraLength:=0;
              end;
            if FEnc then begin
              ExtraLength:=ExtraLength+sizeof(TPKExtraEncField);
              with PkExtraCrypt do begin
                Version:=PkExtraCrypt.Version;
                if Version=2 then CRC:=0 else CRC:=FCRC;
                Method:=PkExtraCrypt.Method;
                end;
              end
            else CRC:=FCRC;
            CommLength:=0;
            DiskStart:=0;
            IntAttr:=0;
            ExtAttr:=Attr;
            end;
          Write(pkd,sizeof(pkd));
          Write(ZfName[1],pkd.FNLength);
          if z64 then Write(lef,ez64l);  // extra fields
          if FEnc then Write(PkExtraCrypt,sizeof(TPKExtraEncField));
          end;
        with pke do begin
          Signatur:=PkEndSignatur;
          ThisDisk:=0;
          StartDisk:=0;
          ThisEntries:=1;
          TotalEntries:=1;
          CommLength:=0;
          end;  // check if file size > 4GB
        if (Position>MaxCardinal) then begin
          // Zip64 end of central directory record
          with pkz do begin
            Signatur:=Pk64RecSignatur;
            Size:=sizeof(TPKZip64CdRecord)-12;
            VersMade:=VersZip64+VersDos;
            VersExtr:=VersZip64+VersDos;
            ThisDisk:=0;
            StartDisk:=0;
            ThisEntries:=1;
            TotalEntries:=1;
            DirSize:=Position-fp;
            Offset:=fp;
            end;
          with pke do begin
            if pkz.DirSize>MaxCardinal then DirSize:=MaxCardinal
            else DirSize:=pkz.DirSize;
            Offset:=MaxCardinal;
            end;
          fp:=Position;
          // Zip64 end of central directory locator
          with pko do begin
            Signatur:=Pk64LocSignatur;
            StartDisk:=0;
            Offset:=fp;
            TotalDisk:=1;
            end;
          Write(pkz,sizeof(TPKZip64CdRecord));
          Write(pko,sizeof(TPKZip64CdLocator));
          end
        else with pke do begin
          DirSize:=Position-fp;
          Offset:=TInt64(fp).Lo;
          end;
        Write(pke,sizeof(pke));
      except
        on EWriteError do ec:=errFileWrite;
        end;
      end
    else ec:=errFileRead;
    if ec=errOK then ec:=CloseFiles(false,FCopyAcl,FTime,Attr) else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

procedure TWinCopyThread.Execute;
var
  ec : integer;
begin
  if scmCopy in FShellCopy then begin
    ec:=ShCopyFile;
    if FVerify and (ec=errOK) then ec:=VerifyFile;
    end
  else if FZipCompress then begin
    ec:=CompressAsZip (FCompress);
    if FVerify and (ec=errOK) then begin
      if FCompress then
      else ;
      end;
    end
  else if FCompress then begin
    ec:=CompressFile;
    if FVerify and (ec=errOK) then ec:=VerifyGzFile;
    end
  else if assigned(FCrypting) then begin
//    DestName:=DestName+CrExt;
    ec:=EncryptFile;
    if FVerify and (ec=errOK) then ec:=VerifyEncFile;
    end
  else begin
    ec:=CopyFile;
    if FVerify and (ec=errOK) then ec:=VerifyFile;
    end;
  EndThread (ec);
  end;

{ TGUnZipThread ------------------------------------------------------------------- }
constructor TGUnZipThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                                  ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,ASuspend,ABufSize,APriority);
  FCrypting:=AEncryption;
  FErrFlag:=errGUnzip;
  end;

{ ------------------------------------------------------------------- }
// read gz header
function TGUnZipThread.GzReadHeader (fSource   : TStream;
                                     var fName : string;
                                     var fTime : cardinal;
                                     var fEnc  : boolean;
                                     var CSize,USize : int64) : integer;
begin
  Result:=GetGzHeader (fSource,fName,fTime,fEnc,CSize,USize);
  end;

function TGUnZipThread.GzSkipHeader (fSource : TStream) : integer;
var
  s : string;
  t : cardinal;
  e : boolean;
  c,u : int64;
begin
  Result:=GetGzHeader(fSource,s,t,e,c,u);
  end;

function TGUnZipThread.GzReadSizes (fSource   : TStream;
                                    var CSize,USize : int64) : integer;
var
  s : string;
  t : cardinal;
  e : boolean;
begin
  Result:=GetGzHeader(fSource,s,t,e,CSize,USize);
  end;

{ ------------------------------------------------------------------- }
// read gz trailer
function TGUnZipThread.GzReadTrailer (fSource : TStream; USize : int64) : integer;
var
  aCrc,FSize : cardinal;
begin
  Result:=errOK;       // o.k.
  try
    fSource.Read(aCrc,4);
    fSource.Read(FSize,4);
    if (FCrc<>aCrc) or (TInt64(USize).Lo<>FSize) or (USize<>UTotal) then begin
      Result:=errError+errFileCheck; exit; // checksum error
      end;
  except
    Result:=errError+errFileWrite; // error
    end;
  end;

{ ------------------------------------------------------------------- }
// uncompress file
function TGUnZipThread.UncompressFile : integer;
var
  FTime         : TFileTimestamps;
  utime         : cardinal;
  ec            : integer;
  FName         : string;
  fp,n,sz       : int64;
  Attr          : word;
  sTmp          : TStream;
  FEnc,tf       : boolean;
  tzn           : string;
begin
  ec:=errOK;
//  if length(SourceName)>MAX_PATH-1 then ec:=errError+errLongPath else
  if not FileExistsUNC (SourceName,false) then ec:=errError+errNotFound;
  if (ec=errOK) and (length(DestName)=0) then ec:=errError+errFileCreate;
  if (ec=errOK) then begin
    FTime:=GetFileTimestamps(SourceName,false);
    Attr:=FileGetAttrUNC(SourceName,false);
    end;
  if ec=errOK then begin
    try
      sSource:=TFileStream.Create(FilenameToUNC(SourceName),fmOpenRead+fmShareDenyNone);
    except
      ec:=errError+errFileOpen;
      end;
    end;
  if ec=errOK then begin
    ec:=GzReadHeader(sSource,FName,utime,FEnc,n,sz);
    if (ec=errOK) then begin
//      if length(FName)>0 then DestName:=DestName+FName
//      else DestName:=DestName+ChangeFileExt(ExtractFilename(SourceName),'');
      // always overwrite destination, reset attributes
      if FOverwrite and FileExistsUNC(DestName,false) then begin
        if FileSetAttrUNC(DestName,faArchive,false)<>0 then ec:=errError+errFileAttr;
        end;
      end;
    if ec=errOK then begin
      try
        sDest:=TFileStream.Create(FilenameToUNC(DestName),fmCreate);
      except
        ec:=errError+errFileCreate;
        end;
      end;
    if (ec=errOK) and FEnc and not assigned(FCrypting) then ec:=errError+errFileDCrypt;
    if ec=errOK then begin
      fp:=sSource.Position;
      if FEnc then begin // decrypt file
        DoProgress(acDecrypt,-sSource.Size,true);
        tf:=sSource.Size>CMaxMemSize;
        if tf then begin
          tzn:=fTmpDir+tmpEncName;
          sTmp:=TFileStream.Create(FilenameToUNC(tzn),fmCreate); // write to disk
          end
        else sTmp:=TMemoryStream.Create;
        with sSource do n:=Size-Position-8;
        if FCrypting.DecryptStream(sSource,sTmp,n,true) then begin
          DoProgress(acUnCompress,-sz,false);
          sTmp.Seek(0,soFromBeginning);
          ec:=GzUncompress(sTmp,sDest,sTmp.size,false);
          end
        else ec:=errError+errFileDCrypt;
        sTmp.Free;
        if tf then DeleteFile(tzn);
        // Ignore trailer
        end
      else begin
        DoProgress(acUnCompress,-sz,true);
        ec:=GzUncompress(sSource,sDest,true);
        if (ec=errOK) then begin
          CTotal:=sSource.Position-fp;
          ec:=GzReadTrailer (sSource,sz);
          end;
        end;
      end;
    if ec=errOK then ec:=CloseFiles (false,FCopyAcl,FTime,Attr) else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

procedure TGUnZipThread.Execute;
begin
  EndThread (UncompressFile);
  end;

{ ------------------------------------------------------------------- }
// Verify two files by comparing bytewise
// AVerifyName = file to be verified, may be compressed or encrypted
// ARefName = original file at source as reference
constructor TCompareThread.Create (const AVerifyName,ARefName : string;  AEncryption : TEncryption;
                                  ACompress,ASuspend : Boolean; ABufSize : integer;
                                  APriority : TThreadPriority);
begin
  inherited Create (AVerifyName,ARefName,AEncryption,ASuspend,ABufSize,APriority);
  FCompress:=ACompress;
  if assigned(FCrypting) then FErrFlag:=errDecrypt
  else if FCompress then FErrFlag:=errGUnZip
  else FErrFlag:=0;
  end;

function TCompareThread.CompareDecryptedFile (fSource,fDest : TStream) : integer;
var
  tf          : boolean;
  ten         : string;
  sTmp        : TStream;
begin
  Result:=errOK;
  tf:=sSource.Size>UMaxMemSize;
  if tf then begin
    ten:=fTmpDir+tmpEncName;
    sTmp:=TFileStream.Create(FilenameToUNC(ten),fmCreate); // write to disk
    end
  else sTmp:=TMemoryStream.Create;
  DoProgress(acDecrypt,-sDest.Size,true);
  if not FCrypting.DecryptStream(sDest,sTmp,sDest.Size,true) then Result:=errError+errFileDCrypt;
  if Result=errOK then Result:=CompareStreams(fSource,sTmp,false);
  sTmp.Free;
  if tf then DeleteFile(ten);
  end;

function TCompareThread.CompareCompressedFile(fSource,fDest : TStream) : integer;
var
  tf          : boolean;
  tzn,ten     : string;
  sETmp,sZTmp : TStream;
  n,sz        : int64;
begin
  Result:=GzReadSizes(fDest,n,sz);
  if Result=errOK then begin
    tf:=fSource.Size>UMaxMemSize;
    if tf then begin
      tzn:=fTmpDir+tmpGzipName;
      sZTmp:=TFileStream.Create(FilenameToUNC(tzn),fmCreate); // write to disk
      end
    else sZTmp:=TMemoryStream.Create;
    if assigned(FCrypting) then begin
      DoProgress(acDecrypt,-sDest.Size,true);
      if tf then begin
        ten:=fTmpDir+tmpEncName;
        sETmp:=TFileStream.Create(FilenameToUNC(ten),fmCreate); // write to disk
        end
      else sETmp:=TMemoryStream.Create;
      with fDest do n:=Size-Position-8;
      if FCrypting.DecryptStream(fDest,sETmp,n,true) then begin
        DoProgress(acUnCompress,-sz,false);
        sETmp.Seek(0,soFromBeginning);
        Result:=GzUncompress(sETmp,sZTmp,sETmp.size,false);
        end
      else Result:=errError+errFileDCrypt;
      sETmp.Free;
      if tf then DeleteFile(ten);
      end
    else begin
      DoProgress(acUnCompress,-sz,true);
      Result:=GzUncompress(fDest,sZTmp,true);
      end;
    if Result=errOK then Result:=CompareStreams(fSource,sZTmp,false);
    sZTmp.Free;
    if tf then DeleteFile(tzn);
    end;
  end;

procedure TCompareThread.Execute;
var
  ec : integer;
begin
  ec:=errOK;
  if not FileExistsUNC (DestName,false) then ec:=errError+errNotFound;
  if (ec=errOK) and not FileExistsUNC (SourceName,false) then ec:=errError2+errNotFound;
  if ec=errOK then begin
    try       // reference file
      sSource:=TFileStream.Create(FilenameToUNC(DestName),fmOpenRead+fmShareDenyNone);
    except
      on EFOpenError do ec:=errError+errFileOpen;
      end;
    end;
  if ec=errOK then begin
    try     // file to be verified
      sDest:=TFileStream.Create(FilenameToUNC(SourceName),fmOpenRead+fmShareDenyNone);
    except
      on EFOpenError do ec:=errError2+errFileOpen;
      end;
    end;
  if ec=errOK then begin
    if FCompress then begin
      ec:=CompareCompressedFile(sSource,sDest);
      end
    else if assigned(FCrypting) then begin
      ec:=CompareDecryptedFile(sSource,sDest);
      end
    else begin
      ec:=CompareStreams(sSource,sDest,true);
      end;
    end;
  ReleaseStreams;
  EndThread (ec);
  end;

{ ------------------------------------------------------------------- }
constructor TPkHeader.Create (const Fn : RawByteString; Ec,Al,Ts,Chk : cardinal;
                              AA,AF,AM,EV,ED,EM : word; Ofs,CSz,USz : int64);
begin
  inherited Create;
  FName:=Fn;
  ErrCode:=Ec;
  Tag:=0;               // for internal use
  DirLength:=Al;        // total length of directory entry (only ReadDirectory)
  Timestamp:=Ts; CRC:=Chk; Attr:=AA; GpFlag:=AF; CompMethod:=AM;
  EncVers:=EV; EncMode:=ED; EncMeth:=EM;
  Offset:=Ofs; CSize:=CSz; USize:=USz;
  Zip64:=(Offset>MaxCardinal) or (CSize>MaxCardinal) or (USize>MaxCardinal);
  IsDir:=(CSz=0) and (AF and not zipUtf8Flag=0);
  end;

// for rebuilding of Zip archives
constructor TPkHeader.Assign (APkHeader : TPkHeader; Ofs : int64);
begin
  inherited Create;
  FName:=APkHeader.FName;
  ErrCode:=APkHeader.ErrCode;
  Timestamp:=APkHeader.TimeStamp;
  CRC:=APkHeader.CRC;
  Attr:=APkHeader.Attr;
  GpFlag:=APkHeader.GpFlag;
  CompMethod:=APkHeader.CompMethod;
  EncVers:=APkHeader.EncVers;
  EncMode:=APkHeader.EncMode;
  EncMeth:=APkHeader.EncMeth;
  Tag:=APkHeader.Tag;
  DirLength:=APkHeader.DirLength;
  Offset:=Ofs;                   // assign new position
  CSize:=APkHeader.CSize;
  USize:=APkHeader.USize;
  Zip64:=APkHeader.Zip64;
  IsDir:=APkHeader.IsDir;
  end;

{ ------------------------------------------------------------------- }
(* generate error message *)
constructor EZipError.Create (const ErrString : string;
                              ErrNR     : integer);
begin
  inherited Create (ErrString);
  ZipError:=ErrNr or ErrZip;
  end;

// =====================================================================
{ TZip  ----------------------------------------------------}
// Common Zip object
constructor TZip.Create (const ZipName : string; const Password : AnsiString;
                         AKeyMode,ABufSize,AMode : integer);
begin
  inherited Create (FilenameToUNC(ZipName),AMode);
  FileList:=TStringList.Create;
  FileList.Sorted:=false;
  if ABufSize>0 then FBufferSize:=ABufSize
  else FBufferSize:=defBufferSize;
  FIsTemp:=false;     // if Zip file is temporary -> no speed limits
  FErrorFlag:=false;
  ZName:=ZipName;
  PBase:='';
  PkExtraCrypt:=defPkExtraCrypt;
  FEncrypt:=false;
  InitEncryption(Password,AKeyMode);
  FPriority:=tpNormal;
  FOnProgress:=nil;
  FOnTimeout:=nil;
  end;

destructor TZip.Destroy;
begin
  if FEncrypt then FEncryption.Free; // release encryption
  FileList.Free;
  inherited Destroy;
  end;

procedure TZip.InitEncryption (const Password : AnsiString; AKeyMode : integer);
begin
  if FEncrypt then FEncryption.Free;
  FEncrypt:=length(Password)>0;
  if FEncrypt then begin   // initialize encryption
    FEncryption:=TEncryption.Create(Password,AKeyMode,FBufferSize);
    PkExtraCrypt.Mode:=FEncryption.KeyMode;
    end
  else FEncryption:=nil;
  end;

procedure TZip.SetProgress (ACallBack : TProgressEvent);
begin
  FOnProgress:=ACallBack;
  if FEncrypt then FEncryption.OnProgress:=ACallBack;
  end;

procedure TZip.CopyToStream(AStream : TStream; SLength : int64);
var
  FBuffer            : array of byte;
  NRead,NWrite,ec,fb : Integer;
begin
  SetLength(FBuffer,FBufferSize);
  ec:=errOK;
  repeat
    if SLength<FBufferSize then fb:=SLength else fb:=FBufferSize;
    try
      NRead:=Read(FBuffer[0],fb);
      if NRead<fb then ec:=errFileRead;   // z.B. wenn "sSource" gelockt ist
    except
      ec:=errFileRead;
      end;
    if ec=errOK then begin
      try
        NWrite:=AStream.Write(FBuffer[0],NRead);
        if NWrite<NRead then ec:=errFileFull; // Ziel-Medium voll
      except
        ec:=errFileWrite;
        end;
      end;
    dec(Slength,NRead);
    until (SLength<=0) or (ec<>errOK);
  if ec<>errOK then
    raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec);
  end;

{ ------------------------------------------------------------------- }
// Read Zip archive
constructor TZipRead.Create (const ZipName : string; const Password : AnsiString;
                             AKeyMode,ABufSize : integer);
begin
  try
    inherited Create (ZipName,Password,AKeyMode,ABufSize,fmOpenRead);
  except
    on EFCreateError do begin
      raise EZipError.Create (Format(rsErrOpening,[ZipName]),errError+errFileOpen);
      end;
    end;
  end;

destructor TZipRead.Destroy;
var
  i : integer;
begin
  with FileList do for i:=0 to Count-1 do Objects[i].Free;
  inherited Destroy;
  end;

// Read central directory
procedure TZipRead.ReadDirectory;
var
  pke   : TPkEndHeader;
  pkd   : TPkDirHeader;
  pkz   : TPKZip64CdRecord;
  pko   : TPKZip64CdLocator;
  ZfName     : RawByteString;
  s,sp       : string;
  usz,csz,fp : int64;
  i,ne,ec,
  el,k      : cardinal;
  pkelf     : TExtraFieldInfo;
  Buffer    : array of byte;
  EV,ED,EM  : word;
begin
  Position:=Size-sizeof(pke);
  Read(pke,sizeof(pke));
  with pke do begin
    if Signatur<>PkEndSignatur then begin
      raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errSignature);
      Exit;
      end;
    if (pke.Offset=MaxCardinal) or (pke.DirSize=MaxCardinal) or (pke.TotalEntries=MaxWord) then begin // Zip64 extension
      Position:=Size-sizeof(pke)-sizeof(pko);  // read Zip64 end of central directory locator
      Read(pko,sizeof(pko));
      if pko.Signatur<>Pk64LocSignatur then begin
        raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errSignature);
        exit;
        end;
      Position:=pko.Offset;                    // read Zip64 end of central directory record
      Read(pkz,sizeof(pkz));
      if pkz.Signatur<>Pk64RecSignatur then begin
        raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errSignature);
        Exit;
        end;
      with pkz do begin
        fp:=Offset;
        ne:=TotalEntries;
        end;
      end
    else with pke do begin
      fp:=Offset;
      ne:=TotalEntries;
      end;
    end;
  Position:=fp;
  bod:=fp;
  sp:='';
  for i:=0 to ne-1 do begin  // read directory entries
    Read(pkd,sizeof(pkd));
    ec:=errOK;
    with pkd do begin
      if Signatur<>PkDirSignatur then begin
        raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errSignature);
        Exit;
        end;
      if not ((VersExtr=VersDef) or (VersExtr=VersZip64)) then ec:=errFormat;
      if not ((Method=zipNoComp) or(Method=zipDeflate) or (Method=zipEncrypted)) then ec:=errFormat;
      fp:=Offset;
      usz:=USize;
      csz:=CSize;
      SetLength(ZfName,FNLength);
      Read(ZfName[1],FNLength);
      if Flag and zipUtf8Flag<>0 then s:=UnixPathToDosPath(UTF8Decode(ZfName))
      else s:=UnixPathToDosPath(StrToAnsi(ZfName));
      if ExtraLength>0 then begin      // read extra fields
        el:=ExtraLength;
        while el>0 do begin  // read extra fields
          Read(pkelf,SizeOf(TExtraFieldInfo));
          with pkelf do begin
            SetLength(Buffer,ESize);
            Read(Buffer[0],ESize);
            k:=0;
            if Tag=ExtraZip64 then begin
              if USz=MaxCardinal then begin
                Move(Buffer[k],USz,8);
                inc(k,8);
                end;
              if (CSz=MaxCardinal) and (ESize>k) then begin
                Move(Buffer[k],CSz,8);
                inc(k,8);
                end;
              if (fp=MaxCardinal) and (ESize>k) then begin
                Move(Buffer[k],fp,8);
                end;
              end
            else if Tag=ExtraCrypt then begin
              Move(Buffer[0],EV,2);  // Version
              ED:=Buffer[4];         // Mode
              Move(Buffer[5],EM,2);  // Methode
              end;
            el:=el-SizeOf(TExtraFieldInfo)-ESize;
            end;
          end;
        end;
      FileList.AddObject(s,TPkHeader.Create(ZfName,ec,sizeof(pkd)+FNLength+ExtraLength,
                         FTimeStamp,CRC,ExtAttr,Flag,Method,EV,ED,EM,fp,CSz,USz));
      end;
    end;
  end;

function TZipRead.Filename(Index : integer) :string;
begin
  with FileList do if (Index>=0) and (Index<Count) then Result:=Strings[Index]
  else Result:='';
  end;

function TZipRead.IsEncrypted(Index : integer) : boolean;
begin
  Result:=false;
  with FileList do if (Index>=0) and (Index<Count) then
    Result:=((Objects[Index] as TPkHeader).GpFlag and zipEncryptFlag)<>0;
  end;

procedure TZipRead.ExtractToFile(Offset : int64; TimeStamp,Attr : integer; const DestFilename : string);
var
  ec   : integer;
  timo : boolean;
begin
  Position:=Offset;
  ZipThread:=TZipReadThread.Create(DestFilename,FEncryption,self,false,not FIsTemp,
                        FBufferSize,FPriority,FOnProgress);
  with ZipThread do begin
    repeat
      Application.ProcessMessages;
      Sleep(10);
      if assigned(FOnTimeout) then timo:=FOnTimeout else timo:=false;
      until Done or timo;
    ec:=ErrorCode and errAllCodes;
    Free;
    end;
  ZipThread:=nil;
  if ec=errOK then begin
    if SetFileAge(DestFileName,TimeStamp)=0 then begin
      if FileSetAttr(DestFileName,Attr,false)>0 then
        raise EZipError.Create (Format(rsErrSetAttr,[DestFileName]),errWarning+errFileAttr);
      end
    else raise EZipError.Create (Format(rsErrTimeStamp,[DestFileName]),errWarning+errFileTS);
    end
  else begin
  // set error flag on write errors
    FErrorFlag:=(ec=errFileWrite) or (ec=errFileFull);
    raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec);
    end;
  end;

procedure TZipRead.ExtractToFile(Index : integer; const DestFilename : string);
begin
  if assigned(FileList) then if (Index>=0) and (Index<FileList.Count) then
        with (FileList.Objects[Index] as TPkHeader) do begin
    if ((GpFlag and zipEncryptFlag)<>0) and not assigned(FEncryption) then begin
      raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errFileDCrypt);
      Exit;
      end;
    ExtractToFile(Offset,TimeStamp,Attr,DestFileName);
    end;
  end;

// Extract to stream (< 4GB, no encryption)
function TZipRead.ExtractToStream(const Filename : string; sDest : TStream) : boolean;
var
  n         : integer;
  ec        : integer;
  timo      : boolean;
begin
  if assigned(FileList) then n:=FileList.IndexOf(Filename) else n:=-1;
  Result:=n>=0;
  if Result then begin
    Position:=(FileList.Objects[n] as TPkHeader).Offset;
    ZipThread:=TZipReadThread.Create(sDest,nil,self,false,not FIsTemp,
                          FBufferSize,FPriority,FOnProgress);
    with ZipThread do begin
      repeat
        Application.ProcessMessages;
        Sleep(10);
        if assigned(FOnTimeout) then timo:=FOnTimeout else timo:=false;
        until Done or timo;
      ec:=ErrorCode and errAllCodes;
      Free;
      end;
    ZipThread:=nil;
    sDest.Position:=0;
    if ec<>errOK then raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec);
    end;
  end;

procedure TZipRead.CompareToFile(Offset : int64; const RefFilename : string);
var
  ec   : integer;
  timo : boolean;
begin
  Position:=Offset;
  ZipThread:=TZipCompareThread.Create(RefFilename,FEncryption,self,false,not FIsTemp,
                           FBufferSize,FPriority,FOnProgress);
  with ZipThread do begin
    repeat
      Application.ProcessMessages;
      Sleep(10);
      if assigned(FOnTimeout) then timo:=FOnTimeout else timo:=false;
      until Done or timo;
    ec:=ErrorCode and errAllCodes;
    Free;
    end;
  ZipThread:=nil;
  if ec<>errOK then begin
  // set error flag on compare errors
    FErrorFlag:=ec=errFileVerify;
    raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec);
    end;
  end;

procedure TZipRead.Cancel;
begin
  if assigned(ZipThread) then with ZipThread do if not Done then CancelThread;
  end;

{ ------------------------------------------------------------------- }
// Create ZIP archive
constructor TZipWrite.Create (const ZipName,BasicDirectory : string;
                              const Password : AnsiString;
                              AKeyMode,ABufSize : integer);
begin
  try
    inherited Create (ZipName,Password,AKeyMode,ABufSize,fmCreate);
    PBase:=IncludeTrailingPathDelimiter(BasicDirectory);
  except
    on EFCreateError do begin
      raise EZipError.Create(Format(rsErrCreating,[ZipName]),errError+errFileCreate);
      end;
    end;
  FOnVerifyProgress:=nil;
  end;

//  Write directory and final block of ZIP archive
destructor TZipWrite.Destroy;
var
  i : integer;
begin
  if assigned(FileList) then with FileList do for i:=0 to Count-1 do Objects[i].Free;
  inherited Destroy;
  end;

procedure TZipWrite.WriteDirectory;
var
  pke : TPkEndHeader;
  pkd : TPkDirHeader;
  pkl : TPkHeader;
  pkz : TPKZip64CdRecord;
  pko : TPKZip64CdLocator;
  fp  : int64;
  ez64l,
  i,n : integer;
  lef : TPKExtraLocalField;
begin
  fp:=Size;
  with FileList do for i:=0 to Count-1 do begin  // central directory structure
    pkl:=Objects[i] as TPkHeader;
    with pkd do begin
      Signatur:=PkDirSignatur;
      Flag:=pkl.GpFlag;
      Method:=pkl.CompMethod;
      FTimeStamp:=pkl.TimeStamp;
      FNLength:=length(pkl.FName);
      if pkl.Zip64 then begin
        VersMade:=VersZip64+VersDos;
        VersExtr:=VersZip64+VersDos;
        with lef do begin
          Tag:=ExtraZip64;
          n:=0;
          if pkl.USize>MaxCardinal then begin
            USize:=MaxCardinal;
            Buffer[n]:=pkl.USize;
            inc(n);
            end
          else USize:=TInt64(pkl.USize).Lo;
          if pkl.CSize>MaxCardinal then begin
            CSize:=MaxCardinal;
            Buffer[n]:=pkl.CSize;
            inc(n);
            end
          else CSize:=TInt64(pkl.CSize).Lo;
          if pkl.Offset>MaxCardinal then begin
            Offset:=MaxCardinal;
            Buffer[n]:=pkl.Offset;
            inc(n);
            end
          else Offset:=TInt64(pkl.Offset).Lo;
          ESize:=8*n;
          ez64l:=4+ESize;
          ExtraLength:=ez64l;
          end;
        end
      else begin
        VersMade:=VersDef+VersDos;
        VersExtr:=VersDef;
        CSize:=TInt64(pkl.CSize).Lo;
        USize:=TInt64(pkl.USize).Lo;
        Offset:=TInt64(pkl.Offset).Lo;
        ExtraLength:=0;
        end;
      if FEncrypt then begin
        ExtraLength:=ExtraLength+sizeof(TPKExtraEncField);
        with PkExtraCrypt do begin
          Version:=pkl.EncVers;
          if Version=2 then CRC:=0 else CRC:=pkl.CRC;
          Method:=pkl.EncMeth;
          end;
        end
      else CRC:=pkl.CRC;
      CommLength:=0;
      DiskStart:=0;
      IntAttr:=0;
      ExtAttr:=pkl.Attr;
      end;
    try
      Write(pkd,sizeof(pkd));
      Write(pkl.FName[1],pkd.FNLength);
      if pkl.Zip64 then Write(lef,ez64l);  // extra fields
      if FEncrypt then Write(PkExtraCrypt,sizeof(TPKExtraEncField));
    except
      raise EZipError.Create (Format(rsErrDirEntry,[ZName]),errError+errFileWrite);
      exit;
      end;
    end;
  with pke do begin
    Signatur:=PkEndSignatur;
    ThisDisk:=0;
    StartDisk:=0;
    if FileList.Count>MaxWord then begin
      ThisEntries:=MaxWord;
      TotalEntries:=MaxWord;
      end
    else begin
      ThisEntries:=FileList.Count;
      TotalEntries:=FileList.Count;
      end;
    CommLength:=0;
    end;  // check if file size > 4GB
  if (Position>MaxCardinal) or (FileList.Count>MaxWord) then begin
    // Zip64 end of central directory record
    with pkz do begin
      Signatur:=Pk64RecSignatur;
      Size:=sizeof(TPKZip64CdRecord)-12;
      VersMade:=VersZip64+VersDos;
      VersExtr:=VersZip64+VersDos;
      ThisDisk:=0;
      StartDisk:=0;
      ThisEntries:=FileList.Count;
      TotalEntries:=FileList.Count;
      DirSize:=Position-fp;
      Offset:=fp;
      end;
    with pke do begin
      if pkz.DirSize>MaxCardinal then DirSize:=MaxCardinal
      else DirSize:=pkz.DirSize;
      Offset:=MaxCardinal;
      end;
    fp:=Position;
    // Zip64 end of central directory locator
    with pko do begin
      Signatur:=Pk64LocSignatur;
      StartDisk:=0;
      Offset:=fp;
      TotalDisk:=1;
      end;
    try
      Write(pkz,sizeof(TPKZip64CdRecord));
      Write(pko,sizeof(TPKZip64CdLocator));
    except
      raise EZipError.Create (Format(rsErrDirEntry,[ZName]),errError+errFileWrite);
      exit;
      end;
    end
  else with pke do begin
    DirSize:=Position-fp;
    Offset:=TInt64(fp).Lo;
    end;
  try
    Write(pke,sizeof(pke));
  except
    raise EZipError.Create (Format(rsErrEndRecord,[ZName]),errError+errFileWrite);
    end;
  end;

procedure TZipWrite.Verify;  // Verify zip archive
var
  i   : integer;
  ok  : boolean;
begin
  ok:=true;
  if assigned(FOnVerifyProgress) then FOnVerifyProgress(acVerify,-FileList.Count,not FIsTemp);
  FOnProgress:=nil;
  with FileList do for i:=0 to Count-1 do begin  // central directory structure
    if assigned(FOnVerifyProgress) then FOnVerifyProgress(acVerify,i,not FIsTemp);
    ok:=ok and VerifyFile(Objects[i] as TPkHeader);
    if not ok then Break;
    end;
  if not ok then begin
    raise EZipError.Create (Format(rsErrVerify,[ZName]),errError+errFileVerify);
    end;
  end;

//  Add directory name to archive
procedure TZipWrite.AddDir (DirName : string);
var
  ZfName     : RawByteString;
  fn         : string;
  ec         : integer;
  LFHeader   : TPKLocalHeader;
  fp         : int64;
begin
  if DirectoryExistsUNC (DirName,false) then begin
    fn:=ExtractRelativePath(PBase,DirName);
    ec:=errOK;
    with LFHeader do begin
      Signatur:=PKLocalSignatur;
      Vers:=VersDef;
      fn:=ExtractRelativePath(PBase,DirName);
      if IsOEMStr(fn) then begin
        Flag:=0; ZfName:=fn;
        ZfName:=StrToOEM(ZfName); // ZIP needs OEM characters (like DOS)
        end
      else begin
        Flag:=zipUtf8Flag;
        ZfName:=UTF8Encode(fn);
        end;
      Method:=zipNoComp;
      FTimeStamp:=DirectoryAge(ExcludeTrailingPathDelimiter(DirName));
      CRC:=0;
      FNLength:=length(ZfName);
      CSize:=0;
      USize:=0;
      ExtraLength:=0;
      end;
    fp:=Position;
    try
      Write(LFHeader,sizeof(TPKLocalHeader));         // local header
      Write(ZfName[1],LFHeader.FNLength);             // filename
      FileList.AddObject(fn,TPkHeader.Create(ZfName,0,0,LfHeader.FTimeStamp,0,
               FileGetAttr(DirName),LfHeader.Flag,LfHeader.Method,0,0,0,fp,0,0));
    except
      on EWriteError do ec:=errFileWrite;
      end;
    if ec<>errOK then begin
      // set error flag on write errors
      FErrorFlag:=(ec=errFileWrite) or (ec=errFileFull);
      raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec);
      end;
    end;
  end;

procedure TZipWrite.RemLastDir;
begin
  with FileList do if Count>0 then begin
    with (Objects[Count-1] as TPkHeader) do if IsDir then begin
      Size:=Offset;
      Free;
      Delete(Count-1);
      end;
    end;
  end;

// Add stream to archive (< 4GB, no encryption)
procedure TZipWrite.AddStream (ASource : TStream; const AName : string; ACompress : boolean = true);
var
  ZfName    : RawByteString;
  ut,ct,fp  : int64;
  rf,fc     : cardinal;
  LFHeader  : TPKLocalHeader;
  ec        : integer;
  timo      : boolean;
begin
  ut:=ASource.Size;
  with LFHeader do begin
    Signatur:=PKLocalSignatur;
    Flag:=0; ZfName:=AName;
    ZfName:=StrToOEM(ZfName); // ZIP needs OEM characters (like DOS)
    Vers:=VersDef;
    Flag:=Flag or zipMaxDeflFlag;
    if (ASource.Size>0) and ACompress then Method:=zipDeflate else Method:=zipNoComp;
    FTimeStamp:=FileTimeToFileAge(DateTimeToFileTime(Now));
    CRC:=0;
    FNLength:=length(ZfName);
    CSize:=0;
    USize:=0;
    ExtraLength:=0;
    end;
  fp:=Position;
  try
    Write(LFHeader,sizeof(TPKLocalHeader));         // local header
    Write(ZfName[1],LFHeader.FNLength);             // filename
    ZipThread:=TZipWriteThread.Create(ASource,nil,self,ACompress,false,not FIsTemp,
                                      FBufferSize,FPriority,FOnProgress);
    with ZipThread do begin
      repeat
        Application.ProcessMessages;
        Sleep(10);
        if assigned(FOnTimeout) then timo:=FOnTimeout else timo:=false;
        until Done or timo;
      ec:=ErrorCode and errAllCodes;
      ut:=UTotal;
      ct:=CTotal;
      fc:=FCrc;
      Free;
      end;
    ZipThread:=nil;
    if ec=errOK then begin
      Position:=fp+14;                // position of crc
      Write(fc,sizeof(cardinal));
      // write 4 byte values
      rf:=TInt64(ct).Lo;
      Write(rf,sizeof(cardinal));
      rf:=TInt64(ut).Lo;
      Write(rf,sizeof(cardinal));
      Seek (0,soEnd);
      FileList.AddObject(AName,TPkHeader.Create(ZfName,0,0,LfHeader.FTimeStamp,fc,
               faArchive,LfHeader.Flag,LfHeader.Method,
               PkExtraCrypt.Version,PkExtraCrypt.Mode,PkExtraCrypt.Method,fp,ct,ut));
      end;
  except
    on EWriteError do ec:=errFileWrite;
    end;
  if ec<>errOK then Size:=fp;    // reset on error
  if ec<>errOK then begin
    // set error flag on write errors
    FErrorFlag:=(ec=errFileWrite) or (ec=errFileFull);
        raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec);
    end;
  end;

// Add file to archive
procedure TZipWrite.AddFile (SourceName : string; ACompress : boolean = true);
var
  ZfName    : RawByteString;
  fn        : string;
  ec        : integer;
  utime     : cardinal;
  LFHeader  : TPKLocalHeader;
  lef       : TPKExtraLocalField;
  FileData  : TWin32FindData;
  ut,ct,fp  : int64;
  z64,zae2,
  timo      : boolean;
  rf,fc,el  : cardinal;
begin
//  if length(SourceName)>MAX_PATH-1 then ec:=errLongPath else
  if FileExistsUNC (SourceName,false) then begin
    if GetFileData(SourceName,FileData) then begin
      utime:=GetFileAge(FileData);
      ut:=GetFileSize(FileData);
      z64:=ut>MaxSize;
      zae2:=ut<AE1Size;
      with LFHeader do begin
        Signatur:=PKLocalSignatur;
        fn:=ExtractRelativePath(PBase,SourceName);
        if IsOEMStr(fn) then begin
          Flag:=0; ZfName:=fn;
          ZfName:=StrToOEM(ZfName); // ZIP needs OEM characters (like DOS)
          end
        else begin
          Flag:=zipUtf8Flag;
          ZfName:=UTF8Encode(fn);
          end;
        if z64 then Vers:=VersZip64 else Vers:=VersDef;
        if FEncrypt then begin
          Flag:=Flag or zipEncryptFlag;
          Method:=zipEncrypted;
          ACompress:=true;         // compress always on encryption
          end
        else begin
          Flag:=Flag or zipMaxDeflFlag;
          if (ut>0) and ACompress then Method:=zipDeflate else Method:=zipNoComp;
          end;
        FTimeStamp:=utime;
        CRC:=0;
        FNLength:=length(ZfName);
        if z64 then begin
          with lef do begin
            Tag:=ExtraZip64;
            ESize:=16;
            Buffer[0]:=0;
            Buffer[1]:=0;
            end;
          CSize:=MaxCardinal;
          USize:=MaxCardinal;
          el:=20;
          ExtraLength:=20;
          end
        else begin
          CSize:=0;
          USize:=0;
          ExtraLength:=0;
          end;
        if FEncrypt then begin
          ExtraLength:=ExtraLength+sizeof(TPKExtraEncField);
          with PkExtraCrypt do begin
            if zae2 then Version:=VendorVersAE2 else Version:=VendorVersAE1;
            if ut>0 then Method:=zipDeflate else Method:=zipNoComp;
            end;
          end;
        end;
      fp:=Position;
      try
        Write(LFHeader,sizeof(TPKLocalHeader));         // local header
        Write(ZfName[1],LFHeader.FNLength);             // filename
        if z64 then Write(lef,el);                      // extra Zip64 field
        if FEncrypt then Write(PkExtraCrypt,sizeof(TPKExtraEncField));  // extra enc field
        ZipThread:=TZipWriteThread.Create(SourceName,FEncryption,self,ACompress,false,not FIsTemp,
                                          FBufferSize,FPriority,FOnProgress);
        with ZipThread do begin
          repeat
            Application.ProcessMessages;
            Sleep(10);
            if assigned(FOnTimeout) then timo:=FOnTimeout else timo:=false;
            until Done or timo;
          ec:=ErrorCode and errAllCodes;
          ut:=UTotal;
          ct:=CTotal;
          fc:=FCrc;
          Free;
          end;
        ZipThread:=nil;
        if ec=errOK then begin
          if FEncrypt and zae2 then Position:=fp+18    // position of compressed size
          else begin                        // Crc if no encryption or AE-1
            Position:=fp+14;                // position of crc
            Write(fc,sizeof(cardinal));
            end;
          if z64 then begin
            // write 8 byte values to extra field
            Position:=fp+sizeof(TPKLocalHeader)+LFHeader.FNLength+4;
            Write (ut,sizeof(int64));
            Write (ct,sizeof(int64));
            end
          else begin
            // write 4 byte values
            rf:=TInt64(ct).Lo;
            Write(rf,sizeof(cardinal));
            rf:=TInt64(ut).Lo;
            Write(rf,sizeof(cardinal));
            end;
          Seek (0,soEnd);
          FileList.AddObject(fn,TPkHeader.Create(ZfName,0,0,utime,fc,
                   FileGetAttr(SourceName),LfHeader.Flag,LfHeader.Method,
                   PkExtraCrypt.Version,PkExtraCrypt.Mode,PkExtraCrypt.Method,fp,ct,ut));
          end;
      except
        on EWriteError do ec:=errFileWrite;
        end;
      if ec<>errOK then Size:=fp;    // reset on error
      end
    else ec:=errFileRead;
    end
  else ec:=errNotFound;
  if ec<>errOK then begin
    // set error flag on write errors
    FErrorFlag:=(ec=errFileWrite) or (ec=errFileFull);
    raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec);
    end;
  end;

{ TPkHeader = class(TObject)
    FName           : string;  // Original filename (OEM)
    TimeStamp,CRC   : cardinal;
    GpFlag,Attr,
    CompMethode     : word;
    Offset,
    CSize,USize     : int64;
    IsDir,
    Zip64           : boolean;
}
function TZipWrite.VerifyFile (PkHeader : TPkHeader) : boolean;
var
  LFHeader  : TPKLocalHeader;
  Buffer    : array of byte;
  timo      : boolean;
begin
  with PkHeader do if CSize>0 then begin
    Position:=Offset;
    // Header berlesen
    Read(LFHeader,sizeof(TPKLocalHeader));         // local header
    SetLength(Buffer,LFHeader.FNLength);
    Read(Buffer[0],LFHeader.FNLength);             // filename
    if LFHeader.ExtraLength>0 then begin
      SetLength(Buffer,LFHeader.ExtraLength);
      Read(Buffer[0],LFHeader.ExtraLength);             // filename
      end;
    ZipThread:=TZipVerifyThread.Create(FEncryption,Crc,self,CSize,LFHeader.Method<>zipNoComp,false,not FIsTemp,
                      FBufferSize,FPriority,FOnProgress);
    with ZipThread do begin
      repeat
        Application.ProcessMessages;
        Sleep(10);
        if assigned(FOnTimeout) then timo:=FOnTimeout else timo:=false;
        until Done or timo;
      Result:=ErrorCode=errOK;
      Free;
      end;
    end
  else Result:=true;
  end;

procedure TZipWrite.Cancel;
begin
  if assigned(ZipThread) then with ZipThread do if not Done then CancelThread;
  end;

procedure TZipWrite.Suspend;
begin
  if assigned(ZipThread) then with ZipThread do if not Done then Suspend;
  end;

procedure TZipWrite.Resume;
begin
  if assigned(ZipThread) then with ZipThread do if not Done then Resume;
  end;

{ TZipWriteThread ------------------------------------------------------------------- }
constructor TZipWriteThread.Create (const AFileName : string; AEncryption : TEncryption;
                               ZStream : TStream; ACompress,ASuspend : Boolean; ASpeedLimit : boolean;
                               ABufSize : integer; APriority : TThreadPriority;
                               AOnProgress : TProgressEvent);
begin
  inherited Create (AFileName,'',AEncryption,ASuspend,ABufSize,APriority);
  FCompress:=ACompress;
  FSpeedLimit:=ASpeedLimit;
  OnProgress:=AOnProgress;
  sDest:=TFileStream(ZStream);
  FErrFlag:=errZip;
  end;

constructor TZipWriteThread.Create (ASource : TStream; AEncryption : TEncryption;
                               ZStream : TStream; ACompress,ASuspend : Boolean; ASpeedLimit : boolean;
                               ABufSize : integer; APriority : TThreadPriority;
                               AOnProgress : TProgressEvent);
begin
  inherited Create ('','',AEncryption,ASuspend,ABufSize,APriority);
  FCompress:=ACompress;
  FSpeedLimit:=ASpeedLimit;
  OnProgress:=AOnProgress;
  sSource:=ASource;
  sDest:=TFileStream(ZStream);
  FErrFlag:=errZip;
  end;

{ ------------------------------------------------------------------- }
// add file to zip file
function TZipWriteThread.AddToZip : integer;
var
  ec : integer;
begin
  ec:=errOK;
  if length(SourceName)=0 then begin
    if not assigned(sSource) then ec:=errError+errStream;
    end
  else
    try
      sSource:=TFileStream.Create(FilenameToUNC(SourceName),fmOpenRead+fmShareDenyNone);
    except
      on EFOpenError do ec:=errError+errFileOpen;
      end;
  if ec=errOK then begin
    ec:=ProcessStream(sSource);
    try FreeAndNil(sSource); except end;
    end
  else if assigned(sSource) then try FreeAndNil(sSource); except end;
  Result:=ec;
  end;

procedure TZipWriteThread.Execute;
begin
  EndThread(AddToZip);
  end;

{ ------------------------------------------------------------------- }
constructor TPkFileRef.Create (Ofs : int64);
begin
  inherited Create;
  Offset:=Ofs;
  end;

{ TZipVerifyThread ------------------------------------------------------------------- }
constructor TZipVerifyThread.Create (AEncryption : TEncryption; ACrc : cardinal;
                   ZStream : TStream; ALength : int64; ACompress,ASuspend : Boolean; ASpeedLimit : boolean;
                   ABufSize : integer; APriority : TThreadPriority; AOnProgress : TProgressEvent);
begin
  inherited Create ('','',AEncryption,ASuspend,ABufSize,APriority);
  FCompress:=ACompress;
  FSpeedLimit:=ASpeedLimit;
  OnProgress:=AOnProgress;
  sDest:=TFileStream(ZStream);
  FLength:=ALength;
  FCrc:=ACrc;
  FErrFlag:=errZip;
  end;

function TZipVerifyThread.VerifyZippedFile : integer;
var
  ec  : integer;
  crc : cardinal;
begin
  Crc:=FCrc;
  FCrc:=Crc32(0,nil,0);
  if assigned(FCrypting) then begin // decrypt file
    if not FCrypting.DecryptStream(sDest,nil,FLength,FSpeedLimit) then ec:=errError+errFileVerify
    else ec:=errOK;
    end
  else begin
    if FCompress then begin
      ec:=GzUncompress(sDest,nil,FLength,FSpeedLimit);
      if (ec=errOK) and (FCrc<>Crc) then ec:=errError+errFileVerify;
      end
    else ec:=errOK;  // to do
    end;
  Result:=ec;
  end;

procedure TZipVerifyThread.Execute;
begin
  EndThread(VerifyZippedFile);
  end;

{ TZipReadThread ------------------------------------------------------------------- }
constructor TZipReadThread.Create (const AFileName : string; AEncryption : TEncryption;
                               ZStream : TStream; ASuspend : Boolean; ASpeedLimit : boolean;
                               ABufSize : integer; APriority : TThreadPriority;
                               AOnProgress : TProgressEvent);
begin
  inherited Create ('',AFileName,AEncryption,ASuspend,ABufSize,APriority);
  FSpeedLimit:=ASpeedLimit;
  OnProgress:=AOnProgress;
  sSource:=TFileStream(ZStream);
  FErrFlag:=errZip;
  end;

constructor TZipReadThread.Create (ADest : TStream; AEncryption : TEncryption;
                    ZStream : TStream; ASuspend,ASpeedLimit : boolean;
                    ABufSize : integer; APriority : TThreadPriority;
                    AOnProgress : TProgressEvent);
begin
  inherited Create ('','',AEncryption,ASuspend,ABufSize,APriority);
  sDest:=ADest;
  FSpeedLimit:=ASpeedLimit;
  OnProgress:=AOnProgress;
  sSource:=TFileStream(ZStream);
  FErrFlag:=errZip;
  end;

function TZipReadThread.ReadZipLocalHeader (var ZipInfo : TZipFileInfo) : boolean;
var
  LFHeader  : TPKLocalHeader;
  el,k      : cardinal;
  pkelf     : TExtraFieldInfo;
  Buffer    : array of byte;
  ZfName    : RawByteString;
begin
  sSource.Read(LFHeader,sizeof(TPKLocalHeader));         // local header
  with LFHeader do if Signatur=PKLocalSignatur then begin
    ZipInfo.TimeStamp:=FTimeStamp;
    ZipInfo.Crc:=Crc;
    ZipInfo.CSize:=CSize;
    ZipInfo.USize:=USize;
    ZipInfo.Flag:=Flag;
    ZipInfo.Method:=Method;
    SetLength(ZfName,FNLength);
    sSource.Read(ZfName[1],FNLength);
    if Flag and zipUtf8Flag<>0 then ZipInfo.Filename:=UnixPathToDosPath(UTF8Decode(ZfName))
    else ZipInfo.Filename:=UnixPathToDosPath(StrToAnsi(ZfName));
    if ExtraLength>0 then begin      // read extra fields
      el:=ExtraLength;
      while el>0 do begin  // read extra fields
        sSource.Read(pkelf,SizeOf(TExtraFieldInfo));
        with pkelf do begin
          SetLength(Buffer,ESize);
          sSource.Read(Buffer[0],ESize);
          k:=0;
          if Tag=ExtraZip64 then begin
            if USize=MaxCardinal then begin
              Move(Buffer[k],ZipInfo.USize,8);
              inc(k,8);
              end;
            if (CSize=MaxCardinal) and (ESize>k) then begin
              Move(Buffer[k],ZipInfo.CSize,8);
              end
            end
          else if Tag=ExtraCrypt then begin
            FCrypting.KeyMode:=Buffer[4];         // Mode
            end;
          el:=el-SizeOf(TExtraFieldInfo)-ESize;
          end;
        end;
      Buffer:=nil;
      end;
    Result:=true;
    end
  else Result:=false;
  end;

function TZipReadThread.CopyToStream (const ZipInfo : TZipFileInfo; fDest : TStream) : integer;
var
  sTmp          : TStream;
  tf            : boolean;
  tzn           : string;
begin
  if (ZipInfo.Flag and zipEncryptFlag)<>0 then begin  // decrypt
    if not assigned(FCrypting) then Result:=errError+errFileDCrypt
    else begin
      DoProgress(acDecrypt,-ZipInfo.CSize,FSpeedLimit);
      tf:=ZipInfo.CSize>CMaxMemSize;
      if tf then begin
        tzn:=fTmpDir+tmpEncName;
        sTmp:=TFileStream.Create(FilenameToUNC(tzn),fmCreate); // write to disk
        end
      else sTmp:=TMemoryStream.Create;
      if FCrypting.DecryptStream(sSource,sTmp,ZipInfo.CSize,FSpeedLimit) then begin
        DoProgress(acUnCompress,-sTmp.Size,FSpeedLimit);
        sTmp.Seek(0,soFromBeginning);
        Result:=GzUncompress(sTmp,fDest,sTmp.size,FSpeedLimit);
        end
      else Result:=errError+errFileDCrypt;
      sTmp.Free;
      if tf then DeleteFile(tzn);
      end;
    end
  else begin
    DoProgress(acUnCompress,-ZipInfo.CSize,FSpeedLimit);
    if ZipInfo.Method=zipDeflate then Result:=GzUncompress(sSource,fDest,ZipInfo.CSize,FSpeedLimit)
    else Result:=CopyStream(sSource,fDest,ZipInfo.USize,FSpeedLimit);
    if (Result=errOK) and (FCrc<>ZipInfo.Crc) then Result:=errError+errFileRead;
    end;
  end;

function TZipReadThread.CopyToFile : integer;
var
  ZipInfo : TZipFileInfo;
begin
  Result:=errOK;
  if (length(DestName)=0) then begin
    if not assigned(sDest) then Result:=errError+errStream;
    end
  else begin
    try
      sDest:=TFileStream.Create(FilenameToUNC(DestName),fmCreate);
    except
      Result:=errError+errFileCreate;
      end;
    end;
  if (Result=errOK) then begin
    if ReadZipLocalHeader(ZipInfo) then Result:=CopyToStream(ZipInfo,sDest)
    else Result:=errError+errFileRead;
    if (length(DestName)>0) then try FreeAndNil(sDest); except end;
    end;
  end;

procedure TZipReadThread.Execute;
begin
  EndThread(CopyToFile);
  end;

function TZipCompareThread.CompareToFile : integer;
var
  tf          : boolean;
  tzn         : string;
  sTmp        : TStream;
  ZipInfo     : TZipFileInfo;
begin
  Result:=errOK;
  if FileExistsUNC (DestName,false) then begin
    try       // reference file
      sDest:=TFileStream.Create(FilenameToUNC(DestName),fmOpenRead+fmShareDenyNone);
    except
      on EFOpenError do Result:=errError+errFileOpen;
      end;
    if Result=errOK then begin
      if ReadZipLocalHeader(ZipInfo) then begin
        tf:=ZipInfo.CSize>CMaxMemSize;
        if tf then begin
          tzn:=fTmpDir+tmpGzipName;
          sTmp:=TFileStream.Create(FilenameToUNC(tzn),fmCreate); // write to disk
          end
        else sTmp:=TMemoryStream.Create;
        Result:=CopyToStream (ZipInfo,sTmp);
        if Result=errOK then Result:=CompareStreams(sDest,sTmp,FSpeedLimit);
        sTmp.Free;
        if tf then DeleteFile(tzn);
        end
      else Result:=errError+errFileRead;
      try FreeAndNil(sDest); except end;
      end;
    end
  else Result:=errError+errNotFound;
  end;

procedure TZipCompareThread.Execute;
begin
  EndThread(CompareToFile);
  end;

end.
